
from PyQt5.QtCore import QObject, Qt, pyqtSignal, QEvent
from PyQt5.QtWidgets import (QApplication, QWidget,QPushButton,QMessageBox,QDesktopWidget,QMainWindow,
                             QVBoxLayout,QHBoxLayout,QGridLayout,QLabel,QRadioButton,QCheckBox,
                             QLineEdit,QGroupBox,QSplitter,QFileDialog, QScrollArea, QInputDialog, QDialog)
from PyQt5.QtGui import QIcon,QFont,QTextCursor,QPixmap,QColor, QFontMetricsF, QKeySequence, QFocusEvent
import qtawesome as qta # https://github.com/spyder-ide/qtawesome

try:
    import parameters
    from Combobox import ComboBox
    from i18n import _
    from widgets import TextEdit, PlainTextEdit
    import utils_ui
    from widgets import statusBar
    from widgets import EditRemarDialog
except ImportError:
    from COMTool import parameters, utils_ui
    from COMTool.i18n import _
    from COMTool.Combobox import ComboBox
    from COMTool.widgets import TextEdit, PlainTextEdit
    from COMTool.widgets import statusBar
    from COMTool.widgets import EditRemarDialog

try:
    from base import Plugin_Base
    import crc
    from protocols import defaultProtocols
except Exception:
    from .base import Plugin_Base
    from . import crc
    from .protocols import defaultProtocols

import os, json, time
from struct import unpack, pack


class Plugin(Plugin_Base):
    '''
        call sequence:
            set vars like hintSignal, hintSignal
            onInit
            onWidget
            onUiInitDone
                send
                onReceived
            getConfig
    '''
    # vars set by caller
    isConnected = lambda : False
    send = lambda x,y:None          # send(data_bytes=None, file_path=None, callback=lambda ok,msg:None)
    hintSignal = None               # hintSignal.emit(type(error, warning, info), title, msg)
    configGlobal = {}
    # other vars
    connParent = "dbg"       # parent id
    connChilds = []          # children ids
    id = "protocol"
    name = _("protocol")

    enabled = False          # user enabled this plugin
    active  = False          # using this plugin

    showReceiveDataSignal = pyqtSignal(str)

    def __init__(self):
        super().__init__()
        if not self.id:
            raise ValueError(f"var id of Plugin {self} should be set")

    def onInit(self, config):
        '''
            init params, DO NOT take too long time in this func
        '''
        default = {
            "version": 1,
            "sendAscii" : True,
            "useCRLF" : False,
            "sendEscape" : True,
            "code": defaultProtocols.copy(),
            "currCode": "default",
            "customSendItems": [
                {
                    "text": "hello",
                    "remark": "hello",
                    "icon": "fa5.hand-paper"
                },{
                    "text": "\\x01\\x03\\x03\\x03\\x03\\x01",
                    "remark": "pre",
                    "icon": "ei.arrow-left",
                    "shortcut": [[16777234, "Left"]]
                },{
                    "text": "\\x01\\x04\\x04\\x04\\x04\\x01",
                    "remark": "next",
                    "icon": "ei.arrow-right",
                    "shortcut": [[16777236, "Right"]]
                },{
                    "text": "\\x01\\x01\\x01\\x01\\x01\\x01",
                    "remark": "ok",
                    "icon": "fa.circle-o",
                    "shortcut": [[16777220, "Return"]]
                },{
                    "text": "\\x01\\x02\\x02\\x02\\x02\\x01",
                    "remark": "ret",
                    "icon": "ei.return-key",
                    "shortcut": [[16777216, "Esc"]]
                }
            ]
        }
        self.config = config
        for k in default:
            if not k in self.config:
                self.config[k] = default[k]
        self.editingDefaults = False
        self.codeGlobals = {"unpack": unpack, "pack": pack, "crc": crc, "encoding": self.configGlobal["encoding"], "print": self.print}
        self.encodeMethod = lambda x:x
        self.decodeMethod = lambda x:x
        self.pressedKeys = []
        self.keyModeClickTime = 0

    def print(self, *args, **kw_args):
        end = "\n"
        start = "[MSG]: "
        if "end" in kw_args:
            end = kw_args["end"]
        if "start" in kw_args:
            start = kw_args["start"]
        string = start  + " ".join(map(str, args)) + end
        self.showReceiveDataSignal.emit(string)


    class ModeButton(QPushButton):
        onFocusIn = pyqtSignal(QFocusEvent)
        onFocusOut = pyqtSignal(QFocusEvent)

        def __init__(self, text, eventFilter, parent = None) -> None:
            super().__init__(text, parent)
            self.installEventFilter(eventFilter)

        def focusInEvent(self, event):
            self.onFocusIn.emit(event)

        def focusOutEvent(self, event):
            self.onFocusOut.emit(event)

    def onWidgetMain(self, parent):
        self.mainWidget = QSplitter(Qt.Vertical)
        self.receiveWidget = TextEdit()
        font = QFont('Menlo,Consolas,Bitstream Vera Sans Mono,Courier New,monospace, Microsoft YaHei', 10)
        self.receiveWidget.setFont(font)
        self.receiveWidget.setLineWrapMode(TextEdit.NoWrap)
        self.clearBtn = QPushButton("")
        self.keyBtneventFilter = self.ModeButtonEventFilter(self.onModeBtnKeyPressEvent, self.onModeBtnKeyReleaseEvent)
        self.keyModeBtn = self.ModeButton(_("Key mode"), self.keyBtneventFilter)
        layoutClearMode = QHBoxLayout()
        layoutClearMode.addWidget(self.clearBtn)
        layoutClearMode.addWidget(self.keyModeBtn)
        clearModeWidget = QWidget()
        clearModeWidget.setLayout(layoutClearMode)
        utils_ui.setButtonIcon(self.clearBtn, "mdi6.broom")
        self.addButton = QPushButton("")
        utils_ui.setButtonIcon(self.addButton, "fa.plus")
        self.customSendScroll = QScrollArea()
        self.customSendScroll.setMinimumHeight(parameters.customSendItemHeight + 20)
        self.customSendScroll.setWidgetResizable(True)
        self.customSendScroll.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.customSendScroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        cutomSendItemsWraper = QWidget()
        self.customSendScroll.setWidget(cutomSendItemsWraper)
        #   wrapper widget
        cutomSendItemsWraper0 = QWidget()
        cutomSendItemsWraper0.setProperty("class", "scrollbar2")
        layout0 = QVBoxLayout()
        layout0.setContentsMargins(0,8,0,0)
        cutomSendItemsWraper0.setLayout(layout0)
        layout0.addWidget(self.customSendScroll)
        customSendItemsLayoutWrapper = QVBoxLayout()
        customSendItemsLayoutWrapper.setContentsMargins(0,0,0,0)
        cutomSendItemsWraper.setLayout(customSendItemsLayoutWrapper)
        # items container
        self.customItems = QWidget()
        self.customSendItemsLayout = QVBoxLayout()
        self.customSendItemsLayout.setContentsMargins(0,0,0,0)
        self.customItems.setLayout(self.customSendItemsLayout)

        customSendItemsLayoutWrapper.addWidget(self.customItems)
        customSendItemsLayoutWrapper.addWidget(self.addButton)
        customSendItemsLayoutWrapper.addStretch(0)

        self.mainWidget.addWidget(self.receiveWidget)
        self.mainWidget.addWidget(clearModeWidget)
        self.mainWidget.addWidget(cutomSendItemsWraper0)
        self.mainWidget.setStretchFactor(0, 2)
        self.mainWidget.setStretchFactor(1, 1)
        self.mainWidget.setStretchFactor(2, 11)
        # event
        self.addButton.clicked.connect(lambda : self.insertSendItem())
        def clearReceived():
            self.receiveWidget.clear();self.statusBar.clear()
        self.clearBtn.clicked.connect(clearReceived)
        def keyModeOn(event):
            self.keyModeBtn.setProperty("class", "deleteBtn")
            utils_ui.updateStyle(self.mainWidget, self.keyModeBtn)
            self.keyModeClickTime = time.time()
            # show all shortcut
            widgets = self.customItems.findChildren(QPushButton, "editRemark")
            for i, w in enumerate(widgets):
                shortcut = "+".join((name for v, name in self.config["customSendItems"][i]["shortcut"]))
                w.setText(shortcut)
                utils_ui.updateStyle(self.mainWidget, w)

        def keyModeOff(event):
            self.keyModeBtn.setProperty("class", "")
            utils_ui.updateStyle(self.mainWidget, self.keyModeBtn)
            self.keyModeClickTime = 0
            # remove all preesed keys even them not release actually, to avoid window swith by ALT+TAB bug
            self.pressedKeys = []
            # hide all shortcut
            widgets = self.customItems.findChildren(QPushButton, "editRemark")
            for w in widgets:
                w.setText("")

        def keyModeTuggle():
            if self.keyModeBtn.property("class") == "deleteBtn":
                if time.time() - self.keyModeClickTime < 0.2:
                    return
                else:
                    self.keyModeBtn.clearFocus()

        self.keyModeBtn.onFocusIn.connect(keyModeOn)
        self.keyModeBtn.onFocusOut.connect(keyModeOff)
        self.keyModeBtn.clicked.connect(keyModeTuggle)
        return self.mainWidget

    def onWidgetSettings(self, parent):
        root = QWidget()
        rootLayout = QVBoxLayout()
        rootLayout.setContentsMargins(0,0,0,0)
        root.setLayout(rootLayout)
        setingGroup = QGroupBox(_("En-decoding settings"))
        layout = QGridLayout()
        setingGroup.setLayout(layout)
        self.codeItems = ComboBox()
        self.codeItemCustomStr = _("Custom, input name")
        self.codeItemLoadDefaultsStr = _("Load defaults")
        self.codeItems.setEditable(True)
        self.codeWidget = PlainTextEdit()
        self.saveCodeBtn = QPushButton(_("Save"))
        self.saveCodeBtn.setEnabled(False)
        self.deleteCodeBtn = QPushButton(_("Delete"))
        btnLayout = QHBoxLayout()
        btnLayout.addWidget(self.saveCodeBtn)
        btnLayout.addWidget(self.deleteCodeBtn)
        layout.addWidget(QLabel(_("Defaults")),0,0,1,1)
        layout.addWidget(self.codeItems,0,1,1,1)
        layout.addWidget(QLabel(_("Code")),1,0,1,1)
        layout.addWidget(self.codeWidget,1,1,1,1)
        layout.addLayout(btnLayout,2,1,1,1)
        serialSendSettingsLayout = QGridLayout()
        sendGroup = QGroupBox(_("Send settings"))
        sendGroup.setLayout(serialSendSettingsLayout)
        self.sendSettingsAscii = QRadioButton(_("ASCII"))
        self.sendSettingsHex = QRadioButton(_("HEX"))
        self.sendSettingsAscii.setToolTip(_("Get send data as visible format, select encoding method at top right corner"))
        self.sendSettingsHex.setToolTip(_("Get send data as hex format, e.g. hex '31 32 33' equal to ascii '123'"))
        self.sendSettingsAscii.setChecked(True)
        self.sendSettingsCRLF = QCheckBox(_("<CRLF>"))
        self.sendSettingsCRLF.setToolTip(_("Select to send \\r\\n instead of \\n"))
        self.sendSettingsCRLF.setChecked(False)
        self.sendSettingsEscape= QCheckBox(_("Escape"))
        self.sendSettingsEscape.setToolTip(_("Enable escape characters support like \\t \\r \\n \\x01 \\001"))
        serialSendSettingsLayout.addWidget(self.sendSettingsAscii,0,0,1,1)
        serialSendSettingsLayout.addWidget(self.sendSettingsHex,0,1,1,1)
        serialSendSettingsLayout.addWidget(self.sendSettingsCRLF, 1, 0, 1, 1)
        serialSendSettingsLayout.addWidget(self.sendSettingsEscape, 1, 1, 1, 1)

        rootLayout.addWidget(sendGroup)
        rootLayout.addWidget(setingGroup)
        # event
        self.sendSettingsAscii.clicked.connect(lambda : self.bindVar(self.sendSettingsAscii, self.config, "sendAscii", bool))
        self.sendSettingsHex.clicked.connect(lambda : self.bindVar(self.sendSettingsHex, self.config, "sendAscii", bool, invert=True))
        self.sendSettingsCRLF.clicked.connect(lambda : self.bindVar(self.sendSettingsCRLF, self.config, "useCRLF", bool))
        self.sendSettingsEscape.clicked.connect(lambda : self.bindVar(self.sendSettingsEscape, self.config, "sendEscape", bool))
        self.saveCodeBtn.clicked.connect(self.saveCode)
        self.deleteCodeBtn.clicked.connect(self.deleteCode)
        self.codeWidget.onSave = self.saveCode
        return root

    def onWidgetStatusBar(self, parent):
        self.statusBar = statusBar(rxTxCount=True)
        return self.statusBar

    def onUiInitDone(self):
        '''
            UI init done, you can update your widget here
            this method runs in UI thread, do not block too long
        '''
        newItems = []
        for item in self.config["customSendItems"]:
            item = self.insertSendItem(item, load=True)
            newItems.append(item)
        self.config["customSendItems"] = newItems
        self.sendSettingsAscii.setChecked(self.config["sendAscii"])
        self.sendSettingsHex.setChecked(not self.config["sendAscii"])
        self.sendSettingsCRLF.setChecked(self.config["useCRLF"])
        self.sendSettingsEscape.setChecked(self.config["sendEscape"])
        self.showReceiveDataSignal.connect(self.showReceivedData)
        # init decoder and encoder
        for k in self.config["code"]:
            self.codeItems.addItem(k)
        self.codeItems.addItem(self.codeItemCustomStr)
        self.codeItems.addItem(self.codeItemLoadDefaultsStr)
        name = self.config["currCode"]
        idx = self.codeItems.findText(self.config["currCode"])
        if idx < 0:
            idx = 0
            name = "default"
        self.codeItems.setCurrentIndex(idx)
        self.selectCode(name)
        self.codeItems.currentIndexChanged.connect(self.onCodeItemChanged) # add here to avoid self.selectCode trigger
        self.codeWidget.textChanged.connect(self.onCodeChanged)

    class ModeButtonEventFilter(QObject):
        def __init__(self, keyPressCb, keyReleaseCb) -> None:
            super().__init__()
            self.keyPressCb = keyPressCb
            self.keyReleaseCb = keyReleaseCb

        def eventFilter(self, obj, evt):
            if evt.type() == QEvent.KeyPress:
                # prevent default key events
                self.keyPressCb(evt)
                return True
            elif evt.type() == QEvent.KeyRelease:
                self.keyReleaseCb(evt)
                return True
            return False

    def onModeBtnKeyPressEvent(self, event):
        # send by shortcut
        key = event.key()
        self.pressedKeys.append(key)
        for item in self.config["customSendItems"]:
            if not "shortcut" in item:
                continue
            shortcut = item["shortcut"]
            if len(shortcut) == len(self.pressedKeys):
                same = True
                for i in range(len(shortcut)):
                    if shortcut[i][0] != self.pressedKeys[i]:
                        same = False
                        break
                if same:
                    self.sendCustomItem(item)

    def onModeBtnKeyReleaseEvent(self, event):
        key = event.key()
        if key in self.pressedKeys:
            self.pressedKeys.remove(key)

    def onKeyPressEvent(self, event):
        pass

    def onKeyReleaseEvent(self, event):
        pass

    def insertSendItem(self, item = None, load = False):
        # itemsNum = self.customSendItemsLayout.count() + 1
        # height = parameters.customSendItemHeight * (itemsNum + 1) + 20
        # topHeight = self.receiveWidget.height() + 100
        # if height + topHeight > self.funcParent.height():
        #     height = self.funcParent.height() - topHeight
        # if height < 0:
        #     height = self.funcParent.height() // 3
        # self.customSendScroll.setMinimumHeight(height)
        if not item:
            item = {"text": "", "remark": None, "icon": None}
        if type(item) == str:
            item = {
                "text": item,
                "remark": None
            }
        text = item["text"]
        remark = item["remark"]
        itemWidget = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        itemWidget.setLayout(layout)
        cmd = QLineEdit(text)
        if remark:
            send = QPushButton(remark)
        else:
            send = QPushButton("")
        if (not "icon" in item) or not item["icon"]:
            item["icon"] = "fa.send"
        if not "shortcut" in item:
            item["shortcut"] = []
        utils_ui.setButtonIcon(send, item["icon"])
        editRemark = QPushButton("")
        editRemark.setObjectName("editRemark")
        utils_ui.setButtonIcon(editRemark, "ei.pencil")
        editRemark.setProperty("class", "remark")
        cmd.setToolTip(text)
        send.setToolTip(text)
        cmd.textChanged.connect(lambda: self.onCustomItemChange(self.customSendItemsLayout.indexOf(itemWidget), cmd, send))
        send.setProperty("class", "smallBtn")
        def sendCustomData(idx):
            self.sendCustomItem(self.config["customSendItems"][idx])
        send.clicked.connect(lambda: sendCustomData(self.customSendItemsLayout.indexOf(itemWidget)))
        delete = QPushButton("")
        utils_ui.setButtonIcon(delete, "fa.close")
        delete.setProperty("class", "deleteBtn")
        layout.addWidget(cmd)
        layout.addWidget(send)
        layout.addWidget(editRemark)
        layout.addWidget(delete)
        delete.clicked.connect(lambda: self.deleteSendItem(self.customSendItemsLayout.indexOf(itemWidget), itemWidget, [send, editRemark, delete]))
        def changeRemark(idx, obj):
            if not "icon" in self.config["customSendItems"][idx]:
                self.config["customSendItems"][idx]["icon"] = None
            shortcut = []
            if "shortcut" in self.config["customSendItems"][idx]:
                shortcut = self.config["customSendItems"][idx]["shortcut"]
            ok, remark, icon, shortcut = EditRemarDialog(obj.text(), self.config["customSendItems"][idx]["icon"], shortcut).exec()
            if ok:
                obj.setText(remark)
                if icon:
                    utils_ui.setButtonIcon(obj, icon)
                else:
                    obj.setIcon(QIcon())
                self.config["customSendItems"][idx]["remark"] = remark
                self.config["customSendItems"][idx]["icon"] = icon
                self.config["customSendItems"][idx]["shortcut"] = shortcut
        editRemark.clicked.connect(lambda: changeRemark(self.customSendItemsLayout.indexOf(itemWidget), send))
        self.customSendItemsLayout.addWidget(itemWidget)
        if not load:
            self.config["customSendItems"].append(item)
        return item

    def deleteSendItem(self, idx, item, iconItems = []):
        for obj in iconItems:
            utils_ui.clearButtonIcon(obj)
        item.setParent(None)
        self.config["customSendItems"].pop(idx)
        # itemsNum = self.customSendItemsLayout.count()
        # height = parameters.customSendItemHeight * (itemsNum + 1) + 20
        # topHeight = self.receiveWidget.height() + 100
        # if height + topHeight > self.funcParent.height():
        #     height = self.funcParent.height() - topHeight
        # self.customSendScroll.setMinimumHeight(height)

    def showReceivedData(self, text: str):
            curScrollValue = self.receiveWidget.verticalScrollBar().value()
            self.receiveWidget.moveCursor(QTextCursor.End)
            endScrollValue = self.receiveWidget.verticalScrollBar().value()
            self.receiveWidget.insertPlainText(text)
            if curScrollValue < endScrollValue:
                self.receiveWidget.verticalScrollBar().setValue(curScrollValue)
            else:
                self.receiveWidget.moveCursor(QTextCursor.End)

    def onReceived(self, data : bytes):
        self.statusBar.addRx(len(data))
        try:
            data = self.decodeMethod(data)
        except Exception as e:
            self.hintSignal.emit("error", _("Error"), _("Run decode error") + " " + str(e))
            return
        if not data:
            return
        for plugin in self.connChilds:
            plugin.onReceived(data)
        if type(data) != str:
            data = self.decodeReceivedData(data, self.configGlobal["encoding"], not self.config["sendAscii"], self.config["sendEscape"])
        self.showReceiveDataSignal.emit(data + "\n")

    def sendData(self, data_bytes=None):
        try:
            data_bytes = self.encodeMethod(data_bytes)
        except Exception as e:
            self.hintSignal.emit("error", _("Error"), _("Run encode error") + " " + str(e))
            return
        if data_bytes:
            self.send(data_bytes, callback=self.onSent)

    def onSent(self, ok, msg, length, path):
        if ok:
            self.statusBar.addTx(length)
        else:
            self.hintSignal.emit("error", _("Error"), _("Send data failed!") + " " + msg)

    def sendCustomItem(self, item):
        text = item["text"]
        dateBytes = self.parseSendData(text, self.configGlobal["encoding"], self.config["useCRLF"], not self.config["sendAscii"], self.config["sendEscape"])
        if dateBytes:
            self.sendData(data_bytes = dateBytes)

    def onCustomItemChange(self, idx, edit, send):
        text = edit.text()
        edit.setToolTip(text)
        send.setToolTip(text)
        self.config["customSendItems"][idx].update({
            "text": text,
            "remark": send.text()
        })

    def onCodeItemChanged(self):
        if self.editingDefaults:
            return
        self.editingDefaults = True
        if self.codeItems.currentText() == self.codeItemCustomStr:
            self.codeItems.clearEditText()
            self.editingDefaults = False
            return
        if self.codeItems.currentText() == self.codeItemLoadDefaultsStr:
            for name in defaultProtocols:
                idx = self.codeItems.findText(name)
                if idx >= 0:
                    self.codeItems.removeItem(idx)
                self.codeItems.insertItem(self.codeItems.count() - 2, name)
                self.config["code"][name] = defaultProtocols[name]
            self.codeItems.setCurrentIndex(0)
            self.selectCode(self.codeItems.currentText())
            self.editingDefaults = False
            return
        # update code from defaults
        self.selectCode(self.codeItems.currentText())
        self.editingDefaults = False

    def selectCode(self, name):
        if name in [self.codeItemCustomStr, self.codeItemLoadDefaultsStr] or not name or not name in self.config["code"]:
            print(f"name {name} invalid")
            return
        self.config["currCode"] = name
        self.codeWidget.clear()
        self.codeWidget.insertPlainText(self.config["code"][name])
        ok, e, d = self.getEnDecodeMethod(self.codeWidget.toPlainText())
        if ok:
            self.encodeMethod = e
            self.decodeMethod = d
        self.saveCodeBtn.setText(_("Save"))
        self.saveCodeBtn.setEnabled(False)

    def getEnDecodeMethod(self, code):
        func = lambda x:x
        try:
            exec(code, self.codeGlobals)
            if (not "decode" in self.codeGlobals) or not "encode" in self.codeGlobals:
                raise ValueError(_("decode and encode method should be in code"))
            return True, self.codeGlobals["encode"], self.codeGlobals["decode"]
        except Exception as e:
            msg = _("Method error") + "\n" + str(e)
            self.hintSignal.emit("error", _("Error"), msg)
        return False, func, func

    def onCodeChanged(self):
        changed = True
        name = self.codeItems.currentText()
        if name in self.config["code"]:
            codeSaved = self.config["code"][name]
            code = self.codeWidget.toPlainText()
            if code == codeSaved:
                changed = False
        if changed:
            self.saveCodeBtn.setText(_("Save") + " *")
            self.saveCodeBtn.setEnabled(True)
        else:
            self.saveCodeBtn.setText(_("Save"))
            self.saveCodeBtn.setEnabled(False)

    def saveCode(self):
        self.editingDefaults = True
        name = self.codeItems.currentText()
        if name in [self.codeItemCustomStr, self.codeItemLoadDefaultsStr] or not name:
            self.hintSignal.emit("warning", _("Warning"), _("Please input code profile name first"))
            self.editingDefaults = False
            return
        idx = self.codeItems.findText(name)
        if idx < 0:
            self.codeItems.insertItem(self.codeItems.count() - 2, name)
        self.editingDefaults = False
        code = self.codeWidget.toPlainText()
        ok, e, d= self.getEnDecodeMethod(code)
        if ok:
            self.encodeMethod = e
            self.decodeMethod = d
            self.config["code"][name] = code
            self.saveCodeBtn.setText(_("Save"))
            self.saveCodeBtn.setEnabled(False)

    def deleteCode(self):
        self.editingDefaults = True
        name = self.codeItems.currentText()
        itemsConfig = [self.codeItemCustomStr, self.codeItemLoadDefaultsStr]
        # QMessageBox.infomation()
        if name in itemsConfig or not name:
            self.hintSignal.emit("warning", _("Warning"), _("Please select a code profile name first to delete"))
            self.editingDefaults = False
            return
        idx = self.codeItems.findText(name)
        if idx < 0:
            self.editingDefaults = False
            return
        self.codeItems.removeItem(idx)
        self.config["code"].pop(name)
        name = list(self.config["code"].keys())
        if len(name) > 0:
            name = name[0]
            self.codeItems.setCurrentText(name)
            self.selectCode(name)
        self.editingDefaults = False





from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import (QApplication, QWidget,QToolTip,QPushButton,QMessageBox,QDesktopWidget,QMainWindow,
                             QVBoxLayout,QHBoxLayout,QGridLayout,QTextEdit,QLabel,QRadioButton,QCheckBox,
                             QLineEdit,QGroupBox,QSplitter)


class Wave(QWidget):

    closed = pyqtSignal()
    updatedisTextRawSignal = pyqtSignal(str)
    buffer = ""

    def __init__(self,parent = None):
        super(Wave,self).__init__(parent)
        self.init()
        self.initEvent()
        self.show()

    def __del__(self):
        pass

    def init(self):
        self.resize(500,400)
        self.mainLayout = QVBoxLayout()
        self.setLayout(self.mainLayout)
        self.disTextRaw = QLabel("0000")
        self.mainLayout.addWidget(self.disTextRaw)

    def initEvent(self):
        self.updatedisTextRawSignal.connect(self.updateTextRaw)

    def displayData(self,bytes):
        try:
            self.buffer += bytes.decode("utf-8")
            end = self.buffer.index("\r\n")
            frame = self.buffer[0:end]
            self.buffer = self.buffer[end+2:]
            self.updatedisTextRawSignal.emit(frame)
            # print("==========\n",frame,",",self.buffer,"\n==========")

        except Exception:
            pass

    def closeEvent(self, event):
        self.closed.emit()
        event.accept()

    def updateTextRaw(self,frame):
        self.disTextRaw.setText(frame)




import sys
import os
try:
    from main2 import main
    from parameters import log
except Exception:
    from COMTool.main2 import main
    from COMTool.parameters import log

def restart_program():
    '''
        restart program, not return
    '''
    python = sys.executable
    log.i("Restarting program, comand: {} {} {}".format(python, python, *sys.argv))
    os.execl(python, python, * sys.argv)

if __name__ == '__main__':
    while 1:
        ret = main()
        if not ret is None:
            break
        restart_program()
    print("-- program exit, code:", ret)
    sys.exit(ret)



from . import dbg
from . import protocol
from .import terminal
from . import graph
# from . import myplugin

pluginClasses = [dbg.Plugin, protocol.Plugin, terminal.Plugin, graph.Plugin]
# pluginClasses.append(myplugin.Plugin)

builtinPlugins = {}
for c in pluginClasses:
    builtinPlugins[c.id] = c







from struct import unpack, pack
import utils

crc16 = crc.crc16

import json

version = 1

# cmd enum
CMD_OK             = 0x00
CMD_ERROR          = 0x7F
CMD_KEY            = 0x01
CMD_APP_LIST       = 0x02
CMD_CUR_APP_INFO   = 0x03
CMD_APP_INFO       = 0x04
CMD_START_APP      = 0x05
CMD_EXIT_APP       = 0x06
CMD_APP_CMD        = 0x08
CMD_APP_SETTING    = 0x09

CMD_RESPONSE_MASK  = 0x80

MSG_TYPE_RAW = 0x00
MSG_TYPE_STR = 0x01
msg_type = ["raw", "str"]


class Base:
    name = ""
    def __init__(self, request=True, body = b''):
        self.request = request
        if body:
            self.decode(body)

    def decode(self, body):
        raise NotImplementedError()

    def encode(self):
        raise NotImplementedError()

class Data_APP_CMD(Base):
    '''
        for cmd CMD_APP_CMD
        | app_len(1B) | app_id | content |
    '''
    name = "app"
    def __init__(self, request=True, body = b'', app_id = "", content = b''):
        self.app_id = app_id
        self.content = content
        self.request = request
        if body:
            self.decode(body)

    def decode(self, body):
        app_len = body[0]
        self.app_id = body[1:app_len + 1].decode("utf-8")
        self.content = body[app_len + 1:]


    def encode(self):
        app_id = self.app_id.encode("utf-8")
        ret = pack("B", len(app_id))
        ret += app_id
        ret += self.content
        return ret

class Data_OK(Base):
    '''
        response
            | cmd(1B) | sub_cmd(1B) | msg_type(1B) | msg |
    '''
    name = "ok"
    def __init__(self, request=False, body = b'', cmd = CMD_OK, sub_cmd = CMD_OK, msg = '', msg_type = MSG_TYPE_STR):
        '''
            @cmd response for cmd
            @sub_cmd response for cub_cmd, for example, cmd is CMD_APP_CMD, and sub_cmd can be 0x01 0x02 etc.
        '''
        self.cmd = cmd
        self.sub_cmd = sub_cmd
        self.msg = msg
        self.msg_type = msg_type
        super().__init__(request, body=body)

    def decode(self, body):
        self.cmd = body[0]
        self.sub_cmd = body[1]
        self.msg_type = body[2]
        if self.msg_type == MSG_TYPE_STR:
            self.msg = body[3:].decode("utf-8")
        else:
            self.msg = body[3:]

    def encode(self):
        body = bytes([self.cmd, self.sub_cmd])
        if self.msg:
            if self.msg_type == MSG_TYPE_RAW:
                body += bytes([MSG_TYPE_RAW])
                body += self.msg
            elif self.msg_type == MSG_TYPE_STR:
                body += bytes([MSG_TYPE_STR])
                body += self.msg.encode("utf-8")
        else:
            body += bytes([MSG_TYPE_RAW])
        return body

    def __str__(self) -> str:
        ret = {
            "type": self.name,
            "msg": str(self.msg) if self.msg else "",
            "cmd": self.cmd,
            "sub_cmd": self.sub_cmd
        }
        return json.dumps(ret, ensure_ascii=False)

class Data_ERROR(Base):
    '''
        response
            | cmd(1B) | sub_cmd(1B) | msg_type(1B) | msg |
    '''
    name = "error"
    def __init__(self, request=False, body = b'', cmd=CMD_ERROR, sub_cmd=CMD_ERROR, msg = '', msg_type = MSG_TYPE_STR):
        self.cmd = cmd
        self.sub_cmd = sub_cmd
        self.msg = msg
        self.msg_type = msg_type
        super().__init__(request, body=body)

    def decode(self, body):
        self.cmd = body[0]
        self.sub_cmd = body[1]
        self.msg_type = body[2]
        if self.msg_type == MSG_TYPE_STR:
            self.msg = body[3:].decode("utf-8")
        else:
            self.msg = body[3:]

    def encode(self):
        body = bytes([self.cmd, self.sub_cmd])
        if self.msg:
            if self.msg_type == MSG_TYPE_RAW:
                body += bytes([MSG_TYPE_RAW])
                body += self.msg
            elif self.msg_type == MSG_TYPE_STR:
                body += bytes([MSG_TYPE_STR])
                body += self.msg.encode("utf-8")
        else:
            body += bytes([MSG_TYPE_RAW])
        return body

    def __str__(self) -> str:
        ret = {
            "type": self.name,
            "msg": str(self.msg) if self.msg else "",
            "cmd": self.cmd,
            "sub_cmd": self.sub_cmd
        }
        return json.dumps(ret, ensure_ascii=False)

class Data_KEY(Base):
    '''
        simulate keyboard input, only have request format:
        |   name  | cmd(1B)     | key(4B) | value(1B) |
        | ------- | ----------- | ------------- | --------- |
        | example | 0x01(fixed) | 0x00000026(LE)    | 0x01(/0x00/0x02) |
        e.g. [0x01, 0x26, 0x00, 0x00, 0x00, 0x01]
        cmd: fixed 0x01
        key: little edian 4 bytes data, value can be:
            38: "up"
            40: "down"
            37: "left"
            39: "right"
            108: "enter"
            27: "esc"
            0x01010101: "ok"
            0x02020202: "ret"
            0x03030303: "pre"
            0x04040404: "next"
        value: key value, can be 0x01(pressed),0x02(long pressed), 0x00(released)
    '''
    name = "key"
    map = {
        38: "up",
        40: "down",
        37: "left",
        39: "right",
        108: "enter",
        27: "esc",
        0x01010101: "ok",
        0x02020202: "ret",
        0x03030303: "pre",
        0x04040404: "next",
    }
    def __init__(self, request = True, body=b'', num=0, k="", v=1):
        self.map2 = {v : k for k, v in self.map.items()}
        self.k = k
        if self.k and self.k in self.map2:
            self.num = self.map2[self.k]
        else:
            self.num = num
        self.v = v   # 0: release, 1: press, 2: long press
        super().__init__(request, body=body)

    def decode(self, body):
        '''
            @body 4 bytes(int32 type) key, little edian, e.g. b'\x73\x00\x00\x00', 1 byte(uint8 type) value: b'\x01'
        '''
        self.num = unpack("i", body[:4])[0]
        if self.num in self.map:
            self.k = self.map[self.num]
        self.v = unpack("B", body[4:5])[0]

    def encode(self):
        p = pack("i", self.num)
        p += pack("B", self.v)
        return p

    def __str__(self) -> str:
        ret = {
            "type": self.name,
            "value": self.v,
            "key number": self.num,
            "key name": self.k
        }
        return json.dumps(ret, ensure_ascii=False)

class Data_APP_List(Base):
    '''
        get app list command,
            request format:
                |   name  | cmd(1B) |
                | ------- | ------- |
                | example | 0x02    |
            response format:
                |   name  | cmd(1B) | number(1B) | app 0 info(id_len(1B),id,name_len(1B),name,brief_len(1B),brief| ... | app n info |
                | ------- | ------- | ---------- | ---------- | ----- |  ------ |
                | example | 0x82    |    0x05    | [4, 'face', 4, 'face', 11, 'face detect'] | ... | ... |
    '''
    name = "app list"
    def __init__(self, request = True,  body=b'', apps = []):
        self.apps = apps
        super().__init__(request, body=body)

    def decode(self, body):
        if self.request:
            return
        self.apps = []
        self.num = unpack("B", body[0:1])[0]
        p = 1
        idx = 0
        while 1:
            id_len = unpack("B", body[p:p+1])[0]
            p += 1
            id = body[p : p + id_len]
            p += id_len
            name_len = unpack("B", body[p:p+1])[0]
            p += 1
            name = body[p : p + name_len]
            p += name_len
            brief_len = unpack("B", body[p:p+1])[0]
            p += 1
            brief = body[p : p + brief_len]
            p += brief_len
            app = {
                "idx": idx,
                "id": id.decode("utf-8"),
                "name": name.decode("utf-8"),
                "brief": brief.decode("utf-8")
            }
            self.apps.append(app)
            idx += 1
            if p >= len(body):
                break

    def encode(self):
        if self.request:
            return b''
        frame = pack("B", len(self.apps))
        for app in self.apps:
            # id
            id = app["id"].encode("utf-8")
            frame += pack("B", len(id))
            frame += id
            # name
            name = app["name"].encode("utf-8")
            frame += pack("B", len(name))
            frame += name
            # brief
            brief = app["brief"].encode("utf-8")
            frame += pack("B", len(brief))
            frame += brief
        return frame

    def __str__(self) -> str:
        if self.request:
            ret = {
                "type": self.name,
                "request": self.request,
            }
        else:
            ret = {
                "type": self.name,
                "request": self.request,
                "app num": self.num,
                "apps": self.apps
            }
        return json.dumps(ret, ensure_ascii=False)

class Data_CUR_APP_Info(Base):
    '''
        get current app info command,
            request format:
                |   name  | cmd(1B) |
                | ------- | ------- |
                | example | 0x03    |
            response format:
                |   name  | cmd(1B) |  idx(1B)                 | app 0 info(id_len(1B),id,name_len(1B),name,brief_len(1B),brief|
                | ------- | ------- | ------------------------ | ---------- |
                | example | 0x83    |    0x02(0xFF means home) | [4, 'face', 4, 'face', 11, 'face detect'] |
    '''
    name = "cur app info"
    def __init__(self, request = True,  body=b'', app_id = "", apps = []):
        self.apps = apps
        self.app_id = app_id
        super().__init__(request, body=body)

    def decode(self, body):
        if self.request:
            return
        self.idx = unpack("B", body[0:1])[0]
        p = 1
        id_len = unpack("B", body[p:p+1])[0]
        p += 1
        id = body[p : p + id_len]
        p += id_len
        name_len = unpack("B", body[p:p+1])[0]
        p += 1
        name = body[p : p + name_len]
        p += name_len
        brief_len = unpack("B", body[p:p+1])[0]
        p += 1
        brief = body[p : p + brief_len]
        p += brief_len
        self.app_id = id.decode("utf-8")
        self.app = {
            "idx": self.idx,
            "id": self.app_id,
            "name": name.decode("utf-8"),
            "brief": brief.decode("utf-8")
        }

    def encode(self):
        if self.request:
            return b''
        idx = -1
        if self.app_id == "home":
            idx = 0xff
            app = {
                "id": "home",
                "name": "home",
                "brief": "home"
            }
        else:
            for i, app in enumerate(self.apps):
                if self.app_id == app["id"]:
                    idx = i
                    break
            if idx < 0:
                return b''
            app = self.apps[idx]
        frame = pack("B", idx)
        # id
        id = app["id"].encode("utf-8")
        frame += pack("B", len(id))
        frame += id
        # name
        name = app["name"].encode("utf-8")
        frame += pack("B", len(name))
        frame += name
        # brief
        brief = app["brief"].encode("utf-8")
        frame += pack("B", len(brief))
        frame += brief
        return frame

    def __str__(self) -> str:
        if self.request:
            ret = {
                "type": self.name,
                "request": self.request,
            }
        else:
            ret = {
                "type": self.name,
                "request": self.request,
                "app": self.app
            }
        return json.dumps(ret, ensure_ascii=False)

class Data_APP_Info(Base):
    '''
        get app info command,
            request format:
                |   name  | cmd(1B) |  idx(1B)                           | app_id(nB) |
                | ------- | ------- | ---------------------------------- | ------ |
                | example | 0x04    | 0x02(set to 0xFF means not set)    | 'face'(if idx set, this can be empty) |
            response format:
                |   name  | cmd(1B) | idx(1B) | app 0 info(id_len(1B),id,name_len(1B),name,brief_len(1B),brief|
                | ------- | ------- | ------- | ---------------- |
                | example | 0x84    |  0x02   | [4, 'face', 4, 'face', 11, 'face detect'] |
    '''
    name = "app info"
    def __init__(self, request = True,  body=b'', idx = 0xff, app_id = "", apps = []):
        self.apps = apps
        self.app_id = app_id
        self.idx = idx
        super().__init__(request, body=body)
        if (not request) and not body:
            idx = -1
            if self.app_id == "home":
                idx = 0xff
                self.app = {
                    "id": "home",
                    "name": "home",
                    "brief": "home"
                }
            else:
                if self.idx != 0xFF:
                    idx = self.idx
                elif not self.app_id:
                    raise ValueError("idx or app_id should be provided")
                else:
                    for i, app in enumerate(self.apps):
                        if self.app_id == app["id"]:
                            idx = i
                            break
                self.app = self.apps[idx]
            self.idx = idx

    def decode(self, body):
        if self.request:
            self.idx = unpack("B", body[0:1])[0]
            self.app_id = body[1:].decode("utf-8")
            return
        self.idx = unpack("B", body[0:1])[0]
        p = 1
        id_len = unpack("B", body[p:p+1])[0]
        p += 1
        id = body[p : p + id_len]
        p += id_len
        name_len = unpack("B", body[p:p+1])[0]
        p += 1
        name = body[p : p + name_len]
        p += name_len
        brief_len = unpack("B", body[p:p+1])[0]
        p += 1
        brief = body[p : p + brief_len]
        p += brief_len
        self.app = {
            "idx": self.idx,
            "id": id.decode("utf-8"),
            "name": name.decode("utf-8"),
            "brief": brief.decode("utf-8")
        }

    def encode(self):
        if self.request:
            return b''
        frame = pack("B", self.idx)
        # id
        id = self.app["id"].encode("utf-8")
        frame += pack("B", len(id))
        frame += id
        # name
        name = self.app["name"].encode("utf-8")
        frame += pack("B", len(name))
        frame += name
        # brief
        brief = self.app["brief"].encode("utf-8")
        frame += pack("B", len(brief))
        frame += brief
        return frame

    def __str__(self) -> str:
        if self.request:
            ret = {
                "type": self.name,
                "request": self.request,
                "idx": self.idx,
                "app_id": self.app_id
            }
        else:
            ret = {
                "type": self.name,
                "request": self.request,
                "idx": self.idx,
                "app": self.app
            }
        return json.dumps(ret, ensure_ascii=False)

class Data_START_APP(Base):
    '''
        start app command, only have request
            request format:
                |   name  | cmd(1B) |  idx(1B)                           | app_id(nB) |
                | ------- | ------- | ---------------------------------- | ------ |
                | example | 0x05    | 0x02(set to 0xFF means not set)    | 'face'(if idx set, this can be empty) |
    '''
    name = "start app"
    def __init__(self, request = True,  body=b''):
        if not request:
            raise ValueError()
        super().__init__(request, body=body)

    def decode(self, body):
        self.idx = unpack("B", body[0:1])[0]
        self.app_id = body[1:].decode("utf-8")

    def encode(self):
        return b''

    def __str__(self) -> str:
        ret = {
            "type": self.name,
            "request": self.request,
            "idx": self.idx,
            "app_id": self.app_id
        }
        return json.dumps(ret, ensure_ascii=False)

class Data_EXIT_APP(Base):
    '''
        exit app command, only have request
            request format:
                |   name  | cmd(1B) |
                | ------- | ------- |
                | example | 0x06    |
    '''
    name = "exit app"
    def __init__(self, request = True,  body=b'',):
        if not request:
            raise ValueError()
        super().__init__(request, body=body)

    def decode(self, body):
        pass

    def encode(self):
        return b''

    def __str__(self) -> str:
        ret = {
            "type": self.name,
            "request": self.request
        }
        return json.dumps(ret, ensure_ascii=False)


cmd_class = {
    CMD_OK: Data_OK,
    CMD_ERROR: Data_ERROR,
    CMD_KEY: Data_KEY,
    CMD_APP_LIST: Data_APP_List,
    CMD_CUR_APP_INFO: Data_CUR_APP_Info,
    CMD_APP_INFO: Data_APP_Info,
    CMD_START_APP: Data_START_APP,
    CMD_EXIT_APP: Data_EXIT_APP,
}


def _decode(raw : bytes):
    '''
        @raw protocol | header(4B) | version(1B) | data len(4B)(cmd + body)  | cmd(1B) | body(nB)     | CRC16_IBM(2B)(all previous) |
             decimal  | 3148663466 | 1           | 5                         | 1       | ...          | 1271                        |
               hex    | 0xBBACCAAA | 0x01        | 0x00000005                | 0x01    | 0x01000073   | 0x04F7                      |
             little edian when convert bytes to integer
             bytes: b'\xAA\xCA\xAC\xBB\x01\x06\x00\x00\x00\x01\x73\x00\x00\x00\x01\xf7\x04'
        @return (request, cmd, data, raw)
                request or response
                cmd
                data
                raw Data remaining after parsing
    '''
    # find valid header and body, and check parity sum
    if not raw:
        return None, None, None, None, raw
    idx = raw.find(b'\xAA\xCA\xAC\xBB')
    if idx < 0:
        return None, None, None, None, raw
    raw = raw[idx:]
    # check frame length, not enough a frame
    if len(raw) < 12:
        return None, None, None, None, raw
    _version = raw[4]
    if _version != version:
        print(f"protocol version is {_version}, but support is {version}")
        return None, None, None, None, raw
    # get length
    length = unpack("I", raw[5:9])[0]
    # get cmd type
    request = False if raw[9] & CMD_RESPONSE_MASK != 0 else True
    cmd = raw[9] & (~CMD_RESPONSE_MASK)
    # check body length, not enough body
    if len(raw) < length + 11:
        return None, None, None, None, raw
    # checksum
    crc0 = unpack("H", raw[9 + length : 11 + length])[0]
    _crc = crc16(raw[:9 + length])
    if crc0 != _crc:
        raw = raw[11+length:]
        return request, None, None, None, None
    # get body
    body = raw[10:length + 9]
    try:
        if cmd in cmd_class:
            data = cmd_class[cmd](request, body=body)
        else:
            data = body
    except Exception as e:
        print("-- [ERROR] decode error" + str(e))
        raw = raw[11+length:]
        return request, None, None, None, raw
    # remove parsed bytes
    raw = raw[11+length:]
    return request, cmd, data, body, raw



###### just change upper #######

def encode(data:bytes):
    cmd = data[0]
    body = data[1:]
    header = b'\xAA\xCA\xAC\xBB'
    data_len = len(body) + 1
    data_len = pack("I", data_len)
    frame = header + bytes([version]) + data_len + bytes([cmd]) + body
    _crc = crc16(frame)
    _crc = pack("H", _crc)
    frame += _crc
    msg = '  [HEX] {}\n  [BYTES]: {}\n'.format(utils.hexlify(data, " ").decode(), str(data)[2:-1])
    print(msg, start = "=> REQUEST\n")
    return frame

def decode(raw:bytes):
    reqeust, cmd, data, body, raw = _decode(raw)
    if cmd is None:
        return b''
    if type(data) == bytes:
        data = bytes([cmd])+data
        data = '<= {}\n  [HEX]: {}\n  [BYTES]:  {}'.format("REQUEST" if reqeust else "RESPONSE", utils.hexlify(data, " ").decode(), str(data)[2:-1])
        return data
    cmd_bytes = bytes([cmd if reqeust else cmd|CMD_RESPONSE_MASK])
    data = '<= {}\n  [HEX]: {}\n  [BYTES]: {}{}\n  [DECODED]: {}\n'.format(
            "REQUEST" if reqeust else "RESPONSE",
            utils.hexlify(cmd_bytes + body, " ").decode(),
            str(cmd_bytes)[2:-1],
            str(body)[2:-1],
            str(data)
    )
    return data




import enum
from PyQt5.QtCore import QObject, Qt
from PyQt5.QtWidgets import (QApplication, QWidget,QPushButton,QMessageBox,QDesktopWidget,QMainWindow,
                             QVBoxLayout,QHBoxLayout,QGridLayout,QTextEdit,QLabel,QRadioButton,QCheckBox,
                             QLineEdit,QGroupBox,QSplitter,QFileDialog, QScrollArea, QListWidget)
try:
    from .base import Plugin_Base
    from Combobox import ComboBox
    from i18n import _
    import utils, parameters
    from conn.base import ConnectionStatus
    from widgets import statusBar
    from plugins.graph_widgets import graphWidgets
except ImportError:
    from COMTool import utils, parameters
    from COMTool.i18n import _
    from COMTool.Combobox import ComboBox
    from COMTool.conn.base import  ConnectionStatus
    from COMTool.widgets import statusBar
    from COMTool.plugins.graph_widgets import graphWidgets
    from COMTool.plugins.base import Plugin_Base


class Plugin(Plugin_Base):
    '''
        call sequence:
            set vars like hintSignal, hintSignal
            onInit
            onWidget
            onUiInitDone
            onActive
                send
                onReceived
            onDel
    '''
    # vars set by caller
    isConnected = lambda o: False
    send = lambda o,x,y:None          # send(data_bytes=None, file_path=None, callback=lambda ok,msg:None), can call in UI thread directly
    ctrlConn = lambda o,k,v:None      # call ctrl func of connection
    hintSignal = None               # hintSignal.emit(type(error, warning, info), title, msg)
    reloadWindowSignal = None       # reloadWindowSignal.emit(title, msg, callback(close or not)), reload window to load new configs
    configGlobal = {}
    # other vars
    connParent = "main"      # parent id
    connChilds = []          # children ids
    id = "graph"
    name = _("Graph")

    enabled = False          # user enabled this plugin
    active  = False          # using this plugin

    help = '{}<br><br>{}<br><h2>Python</h2><br><pre>{}</pre><p><p>{}<br>{}</p><p>{}</p></p><br><h2>C/C++</h2><br><pre>{}</pre>'.format(
        _("Double click graph item to add a graph widget"), _("line chart plot protocol:"),
'''
from COMTool.plugins import graph_protocol

# For ASCII protocol("binary protocol" not checked)
frame = graph_protocol.plot_pack(name, x, y, binary = False)

# For binary protocol("binary protocol" checked)
frame = graph_protocol.plot_pack(name, x, y, header= b'\\xAA\\xCC\\xEE\\xBB')
''',
        _("Full demo see:"),
        '<a href="https://github.com/Neutree/COMTool/tree/master/tool/send_curve_demo.py">https://github.com/Neutree/COMTool/tree/master/tool/send_curve_demo.py</a>',
        _("Install comtool by <code>pip install comtool</code> first"),
'''

/*******'''+ _('For ASCII protocol("binary protocol" not checked)') + ''' *******/
/**
 * $[line name],[x],[y]&lt;,checksum&gt;\\n
 *   ''' + _('"$" means start of frame, end with "\\n" "," means separator') + ''',
 *   ''' + _('checksum is optional, checksum is sum of all bytes in frame except ",checksum".') + '''
 *   ''' + _('e.g.') + '''
 *     "$roll,1.0,2.0\\n"
 *     "$pitch,1.0,2.0\\r\\n"
 *     "$pitch,1.0,2.0,179\\n" (179 = sum(b"$pitch,1.0,2.0") % 256)
 */
int plot_pack_ascii(uint8_t *buff, int buff_len, const char *name, float x, float y)
{
    snprintf(buff, buff_len, "$%s,%f,%f", name, x, y);
    // add checksum
    int sum = 0;
    for (int i = 0; i &lt; strlen(buff); i++)
    {
        sum += buff[i];
    }
    snprintf(buff + strlen(buff), buff_len - strlen(buff), ",%d\\n", sum & 0xFF);
    return strlen(buff);
}

uint8_t buff[64];
double x = 1.0, y = 2.0;
int len = plot_pack_ascii(buff, sizeof(buff), header, sizeof(header), "data1", x, y);
send_bytes(buff, len);
/*****************************************************************/


/******* ''' + _('For binary protocol("binary protocol" checked)') + ''' *******/
int plot_pack_binary(uint8_t *buff, int buff_len,
               uint8_t *header, int header_len,
               char *name,
               double x, double y)
{
    uint8_t len = (uint8_t)strlen(name);
    int actual_len = header_len + 1 + len + 8 + 8 + 1;
    assert(actual_len &lt;= buff_len);

    memcpy(buff, header, header_len);
    buff[header_len] = len;
    memcpy(buff + 5, name, len);
    memcpy(buff + 5 + len, &x, 8);
    memcpy(buff + 5 + len + 8, &y, 8);
    int sum = 0;
    for (int i = 0; i &lt; header_len+1+len+8+8; i++)
    {
        sum += buff[i];
    }
    buff[header_len+1+len+8+8] = (uint8_t)(sum & 0xff);
    return header_len+1+len+8+8+1;
}

uint8_t buff[64];
uint8_t header[] = {0xAA, 0xCC, 0xEE, 0xBB};
double x = 1.0, y = 2.0;
int len = plot_pack_binary(buff, sizeof(buff), header, sizeof(header), "data1", x, y);
send_bytes(buff, len);
/*****************************************************************/


''')

    def __init__(self):
        super().__init__()
        if not self.id:
            raise ValueError(f"var id of Plugin {self} should be set")

    def onInit(self, config):
        '''
            init params, DO NOT take too long time in this func
            @config dict type, just change this var's content,
                               when program exit, this config will be auto save to config file
        '''
        self.config = config
        default = {
            "version": 1,
            "graphWidgets": [
                # {
                #     "id": "plot",
                #     "config": {}
                # }
            ]
        }
        for k in default:
            if not k in self.config:
                self.config[k] = default[k]
        self.widgets = []

    def onDel(self):
        pass

    def onWidgetMain(self, parent):
        '''
            main widget, just return a QWidget object
        '''
        widget = QWidget()
        widget.setProperty("class", "scrollbar2")
        layout = QVBoxLayout(widget)
        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        layout.addWidget(scroll)
        widget2 = QWidget()
        scroll.setWidget(widget2)
        self.widgetsLayout = QVBoxLayout()
        widget2.setLayout(self.widgetsLayout)
        widget.resize(600, 400)
        # load graph widgets
        for item in self.config["graphWidgets"]:
            if not item["id"] in graphWidgets:
                continue
            c = graphWidgets[item["id"]]
            w = c(hintSignal = self.hintSignal, rmCallback = self.rmWidgetFromMain, send=self.sendData, config=item["config"])
            self.widgets.append(w)
            self.widgetsLayout.addWidget(w)
        return widget

    def onWidgetSettings(self, parent):
        '''
            setting widget, just return a QWidget object or None
        '''
        itemList = QListWidget()
        for k,v in graphWidgets.items():
            itemList.addItem(k)
        itemList.setToolTip(_("Double click to add a graph widget"))
        itemList.setCurrentRow(0)
        itemList.itemDoubleClicked.connect(self.addWidgetToMain)
        return itemList

    def addWidgetToMain(self, item):
        for k, c in graphWidgets.items():
            if k == item.text():
                config = {
                    "id": c.id,
                    "config": {}
                }
                w = c(hintSignal = self.hintSignal, rmCallback = self.rmWidgetFromMain, send=self.sendData, config=config["config"])
                self.widgets.append(w)
                self.widgetsLayout.addWidget(w)
                self.config["graphWidgets"].append(config)

    def rmWidgetFromMain(self, widget):
        self.widgetsLayout.removeWidget(widget)
        for item in self.config["graphWidgets"]:
            if id(item["config"]) == id(widget.config):
                self.config["graphWidgets"].remove(item)
                break
        widget.deleteLater()
        self.widgets.remove(widget)

    def onWidgetFunctional(self, parent):
        '''
            functional widget, just return a QWidget object or None
        '''
        button = QPushButton(_("Clear count"))
        button.clicked.connect(self.clearCount)
        return button

    def onWidgetStatusBar(self, parent):
        self.statusBar = statusBar(rxTxCount=True)
        return self.statusBar

    def clearCount(self):
        self.statusBar.clear()

    def onReceived(self, data : bytes):
        '''
            call in receive thread, not UI thread
        '''
        self.statusBar.addRx(len(data))
        for w in self.widgets:
            w.onData(data)

    def sendData(self, data:bytes):
        '''
            send data, chidren call send will invoke this function
            if you send data in this plugin, you can directly call self.send
        '''
        self.send(data, callback=self.onSent)

    def onSent(self, ok, msg, length, path):
        if ok:
            self.statusBar.addTx(length)
        else:
            self.hintSignal.emit("error", _("Error"), _("Send data failed!") + " " + msg)

    def onKeyPressEvent(self, event):
        for w in self.widgets:
            w.onKeyPressEvent(event)

    def onKeyReleaseEvent(self, event):
        for w in self.widgets:
            w.onKeyReleaseEvent(event)

    def onUiInitDone(self):
        '''
            UI init done, you can update your widget here
            this method runs in UI thread, do not block too long
        '''
        pass

    def onActive(self):
        '''
            plugin active
        '''
        pass



from struct import pack, unpack

def plot_pack(name:str, x:float, y:float, header= b'\xAA\xCC\xEE\xBB', binary = True):
    name = name.encode()
    if binary:
        f = header
        f+= pack("B", len(name ))
        f+=name
        f += pack("d", x)
        f += pack("d", y)
        f += pack("B", sum(f)%256)
    else:
        f = "${},{},{}".format(name, x, y).encode()
        checksum = sum(f) & 0xFF
        f += b',{:d}\n'.format(checksum)
    return f



from audioop import add
import ctypes
from PyQt5.QtCore import pyqtSignal, QPoint, Qt, QEvent, QObject
from PyQt5.QtWidgets import (QWidget, QHBoxLayout, QVBoxLayout, QLabel, QStyleOption, QStyle, QPushButton, QTextEdit,
                            QPlainTextEdit, QMainWindow, QComboBox, QListView, QTabWidget, QStackedWidget, QListWidget,
                            QGridLayout, QLineEdit, QDialog, QScrollArea)
from PyQt5.QtGui import QIcon, QPixmap, QPainter, QMouseEvent, QColor, QKeyEvent, QHideEvent, QKeySequence
import qtawesome as qta # https://github.com/spyder-ide/qtawesome
import os, sys

try:
    import utils_ui
    from Combobox import ComboBox
    from i18n import _
    import parameters
    from qta_icon_browser import selectIcon
except Exception:
    from COMTool import utils_ui
    from COMTool.Combobox import ComboBox
    from COMTool.i18n import _
    from COMTool import parameters
    from COMTool.qta_icon_browser import selectIcon


class TitleBar(QWidget):
    def __init__(self, parent, icon=None, title="", height=35,
                        btnIcons = None,
                        brothers=[],
                        widgets=[[], []]
                ) -> None:
        super().__init__()
        self._height = height
        self.parent = parent
        if not btnIcons:
            btnIcons = [
                "mdi.window-minimize",
                ["mdi.window-maximize", "mdi.window-restore"],
                "mdi.window-close",
                ["ph.push-pin-bold", "ph.push-pin-fill"]
            ]
        self.btnIcons = btnIcons
        layout = QHBoxLayout()
        if brothers:
            rootLayout = QVBoxLayout()
            rootLayout.setContentsMargins(0,0,0,0)
            rootLayout.setSpacing(0)
            widget = QWidget()
            widget.setProperty("class", "TitleBar")
            widget.setLayout(layout)
            widget.setFixedHeight(height)
            rootLayout.addWidget(widget)
            for w in brothers:
                rootLayout.addWidget(w)
                self._height += w.height()
            self.setLayout(rootLayout)
        else:
            self.setLayout(layout)
        self.setFixedHeight(self._height)
        layout.setContentsMargins(0,0,0,0)
        layout.setSpacing(0)
        iconWidget = None
        if icon and os.path.exists(icon):
            iconWidget = QLabel()
            iconWidget.setPixmap(QPixmap(icon).scaled(height, height))
            iconWidget.setProperty("class", "icon")
        self.min = QPushButton("")
        self.max = QPushButton("")
        self.close = QPushButton("")
        self.top = QPushButton("")
        utils_ui.setButtonIcon(self.min, btnIcons[0])
        utils_ui.setButtonIcon(self.max, btnIcons[1][0])
        utils_ui.setButtonIcon(self.close, btnIcons[2])
        utils_ui.setButtonIcon(self.top, btnIcons[3][0])
        self.title = QLabel(title)
        widgets_l, widgets_r = widgets
        if sys.platform.startswith("darwin"):
            layout.addWidget(self.close)
            layout.addWidget(self.min)
            layout.addWidget(self.max)
            for w in widgets_r:
                layout.addWidget(w)
            layout.addStretch(0)
            if iconWidget:
                layout.addWidget(iconWidget)
            layout.addWidget(self.title)
            layout.addStretch(0)
            layout.addWidget(self.top)
            for w in widgets_l:
                layout.addWidget(w)
        else:
            if iconWidget:
                layout.addWidget(iconWidget)
            layout.addWidget(self.title)
            layout.addWidget(self.top)
            for w in widgets_l:
                layout.addWidget(w)
            layout.addStretch(0)
            for w in widgets_r:
                layout.addWidget(w)
            layout.addWidget(self.min)
            layout.addWidget(self.max)
            layout.addWidget(self.close)
        self.min.setFixedHeight(height)
        self.max.setFixedHeight(height)
        self.close.setFixedHeight(height)
        self.top.setFixedHeight(height)

        self.min.setMinimumWidth(height)
        self.max.setMinimumWidth(height)
        self.close.setMinimumWidth(height)
        self.top.setMinimumWidth(height)

        self.min.setProperty("class", "min")
        self.max.setProperty("class", "max")
        self.close.setProperty("class", "close")
        self.title.setProperty("class", "title")
        self.top.setProperty("class", "top")
        self.close.clicked.connect(lambda : parent.close())
        self.max.clicked.connect(lambda : self.onSetMaximized(fromMaxBtn=True))
        self.min.clicked.connect(lambda : parent.setWindowState(Qt.WindowNoState) if parent.windowState() == Qt.WindowMinimized else parent.setWindowState(Qt.WindowMinimized))
        self.top.clicked.connect(lambda : self.onSetTop())
        self.setProperty("class", "TitleBar")

    def mouseDoubleClickEvent(self, event):
        if event.buttons() == Qt.LeftButton:
            self.onSetMaximized()

    def onSetMaximized(self, isMax = None, fromMaxBtn=False, fullScreen = False):
        if not isMax is None:
            if isMax:
                utils_ui.setButtonIcon(self.max, self.btnIcons[1][1])
            else:
                utils_ui.setButtonIcon(self.max, self.btnIcons[1][0])
            return
        status = Qt.WindowNoState
        if fullScreen:
            if self.parent.windowState() != Qt.WindowFullScreen:
                status = Qt.WindowFullScreen
        elif self.parent.windowState() == Qt.WindowNoState:
            if fromMaxBtn and sys.platform.startswith("darwin"): # mac max button to full screen
                status = Qt.WindowFullScreen
            else:
                status = Qt.WindowMaximized
        if status == Qt.WindowNoState:
            utils_ui.setButtonIcon(self.max, self.btnIcons[1][0])
        else:
            utils_ui.setButtonIcon(self.max, self.btnIcons[1][1])
        self.parent.setWindowState(status)
        if status == Qt.WindowFullScreen:
            self.hide()
        else:
            self.show()

    def onSetTop(self, on = None):
        flags = self.parent.windowFlags()
        needShow = self.parent.isVisible()
        def _on(flags):
            flags |= Qt.WindowStaysOnTopHint
            self.parent.setWindowFlags(flags)
            utils_ui.setButtonIcon(self.top, self.btnIcons[3][1])
            self.top.setProperty("class", "topActive")
        def _off(flags):
            flags &=  (~Qt.WindowStaysOnTopHint)
            self.parent.setWindowFlags(flags)
            utils_ui.setButtonIcon(self.top, self.btnIcons[3][0])
            self.top.setProperty("class", "top")
        oldOn = flags & Qt.WindowStaysOnTopHint
        if on is None:
            on = False if oldOn else True
        if on:
            if not oldOn:
                _on(flags)
        else:
            if oldOn:
                _off(flags)
        self.style().unpolish(self.top)
        self.style().polish(self.top)
        self.update()
        if needShow:
            self.parent.show()

    def paintEvent(self, event):
        opt = QStyleOption()
        opt.initFrom(self)
        p = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, p, self)

    def setTitle(self, title):
        self.title.setText(title)

class EventFilter(QObject):
    Margins = 5  # 边缘边距
    windows = []
    _readyToMove = False
    _moving = False
    _resizeCursor = False

    def listenWindow(self, window):
        self.windows.append(window)

    def unlistenWindow(self, window):
        self.windows.remove(window)

    def _get_edges(self, pos, width, height, offset=0):
        edge = 0
        x, y = pos.x(), pos.y()

        if y <= self.Margins - offset and y >= 0:
            edge |= Qt.TopEdge
        if x <= self.Margins - offset and x >= 0:
            edge |= Qt.LeftEdge
        if x >= width - self.Margins + offset and x < width:
            edge |= Qt.RightEdge
        if y >= height - self.Margins + offset and y < height:
            edge |= Qt.BottomEdge
        return edge

    def _get_cursor(self, edges):
        if edges == Qt.LeftEdge | Qt.TopEdge or edges == Qt.RightEdge | Qt.BottomEdge:
            self._resizeCursor = True
            return Qt.SizeFDiagCursor
        elif edges == Qt.RightEdge | Qt.TopEdge or edges == Qt.LeftEdge | Qt.BottomEdge:
            self._resizeCursor = True
            return Qt.SizeBDiagCursor
        elif edges == Qt.LeftEdge or edges == Qt.RightEdge:
            self._resizeCursor = True
            return Qt.SizeHorCursor
        elif edges == Qt.TopEdge or edges == Qt.BottomEdge:
            self._resizeCursor = True
            return Qt.SizeVerCursor
        if self._resizeCursor:
            self._resizeCursor = False
            return Qt.ArrowCursor
        return None

    def moveOrResize(self, window, pos, width, height):
        edges = self._get_edges(pos, width, height)
        if edges:
            if window.windowState() == Qt.WindowNoState:
                window.startSystemResize(edges)
        else:
            if window.windowState() != Qt.WindowFullScreen:
                window.startSystemMove()

    def eventFilter(self, obj, event):
        # print(obj, event.type(), obj.isWindowType(), QEvent.MouseMove)
        if obj.isWindowType():
            # top window 处理光标样式
            if event.type() == QEvent.MouseMove and obj.windowState() == Qt.WindowNoState:
                cursor = self._get_cursor(self._get_edges(event.pos(), obj.width(), obj.height(), offset=1))
                if not cursor is None:
                    obj.setCursor(cursor)
            if event.type() == QEvent.TouchUpdate and not self._moving:
                self._moving = True
                self.moveOrResize(obj, event.pos(), obj.width(), obj.height())
        elif isinstance(event, QMouseEvent):
            if obj in self.windows:
                if event.button() == Qt.LeftButton :
                    if event.type() == QEvent.MouseButtonPress:
                        self._readyToMove = True
                    # elif event.type() == QEvent.MouseButtonDblClick:
                    #     print(obj, event.type(), event)
                elif event.type() == QEvent.MouseMove and self._readyToMove and not self._moving:
                    self._moving = True
                    self.moveOrResize(obj.windowHandle(), event.pos(), obj.width(), obj.height())
        if event.type() == QEvent.MouseButtonRelease or event.type() == QEvent.Move:
            self._readyToMove = False
            self._moving = False
        return False


class CustomTitleBarWindowMixin:
    def __init__(self, titleBar = None, init = False, title = "", icon = os.path.join(parameters.assetsDir, "logo.png")):
        if not init:
            return
        isQMainWindow = False
        for base in self.__class__.__bases__:
            if base == QMainWindow:
                isQMainWindow = True
                break
        if isQMainWindow:
            self.root = QWidget()
            self.setCentralWidget(self.root)
        else:
            self.root = self
        self.root.setProperty("class", "customTilebarWindow")
        self.rootLayout = QVBoxLayout()
        # title bar
        if titleBar:
            self.titleBar = titleBar
        else:
            self.titleBar = TitleBar(self, icon = icon, title=title, height=35)
        # setWindowIcon
        self.setWindowIcon(QIcon(icon))
        if sys.platform == "win32":
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID("comtool") # for add taskbar icon
        self.contentWidget = QWidget()
        self.rootLayout.addWidget(self.titleBar)
        self.rootLayout.addWidget(self.contentWidget)
        self.root.setLayout(self.rootLayout)
        self.rootLayout.setContentsMargins(0, 0, 0, 0) # padding
        self.root.setMouseTracking(True)
        self.titleBar.setMouseTracking(True)
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowMinimizeButtonHint)
        if sys.platform == "win32":
            try:
                from win32_utils import addShadowEffect
            except Exception:
                from COMTool.win32_utils import addShadowEffect
            addShadowEffect(self.winId())
        self.init_vars()

    def changeEvent(self, event):
        # super(CustomTitleBarWindowMixin, self).changeEvent(event)
        self.titleBar.onSetMaximized(isMax = self.isMaximized())

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_F11:
            self.titleBar.onSetMaximized(fullScreen=True)

    def keyReleaseEvent(self,event):
        pass

    # def paintEvent(self, event):
    #     # 透明背景但是需要留下一个透明度用于鼠标捕获
    #     painter = QPainter(self)
    #     painter.fillRect(self.rect(), QColor(255, 255, 255, 1))

    def init_vars(self):
        self._move_drag = False
        self._corner_drag = False
        self._bottom_drag = False
        self._right_drag = False
        self._padding = 6
        self.mPos = None

    # def resizeEvent(self, QResizeEvent):
    #     self._right_rect = [QPoint(x, y) for x in range(self.width() - self._padding + 1, self.width())
    #             for y in range(1, self.height() - self._padding)]
    #     self._bottom_rect = [QPoint(x, y) for x in range(1, self.width() - self._padding)
    #             for y in range(self.height() - self._padding + 1, self.height())]
    #     self._corner_rect = [QPoint(x, y) for x in range(self.width() - self._padding, self.width() + 1)
    #                 for y in range(self.height() - self._padding, self.height() + 1)]

    # def mousePressEvent(self, event):
    #     if event.button() == Qt.LeftButton and not self.mPos:
    #         self.mPos = event.pos()
    #         event.accept()
    #     return

    #     if (event.button() == Qt.LeftButton) and (event.pos() in self._corner_rect):
    #         self._corner_drag = True
    #         event.accept()
    #     elif (event.button() == Qt.LeftButton) and (event.pos() in self._right_rect):
    #         self._right_drag = True
    #         event.accept()
    #     elif (event.button() == Qt.LeftButton) and (event.pos() in self._bottom_rect):
    #         self._bottom_drag = True
    #         event.accept()
    #     # elif (event.button() == Qt.LeftButton) and (event.y() < self.titleBar.height()):
    #     #     self._move_drag = True
    #     #     self.move_DragPosition = event.globalPos() - self.pos()
    #     #     event.accept()

    # def mouseMoveEvent(self, event): # QMouseEvent
    #     if event.buttons() == Qt.LeftButton and self.mPos:
    #         pos = self.mapToGlobal(event.pos() - self.mPos)
    #         if self.windowState() == Qt.WindowMaximized or self.windowState() == Qt.WindowFullScreen:
    #             return
    #         self.move(pos)
    #         event.accept()
    #     return

    #     if QMouseEvent.pos() in self._corner_rect:
    #         self.setCursor(Qt.SizeFDiagCursor)
    #     elif QMouseEvent.pos() in self._bottom_rect:
    #         self.setCursor(Qt.SizeVerCursor)
    #     elif QMouseEvent.pos() in self._right_rect:
    #         self.setCursor(Qt.SizeHorCursor)
    #     else:
    #         self.setCursor(Qt.ArrowCursor)
    #     if Qt.LeftButton and self._right_drag:
    #         self.resize(QMouseEvent.pos().x(), self.height())
    #         QMouseEvent.accept()
    #     elif Qt.LeftButton and self._bottom_drag:
    #         self.resize(self.width(), QMouseEvent.pos().y())
    #         QMouseEvent.accept()
    #     elif Qt.LeftButton and self._corner_drag:
    #         self.resize(QMouseEvent.pos().x(), QMouseEvent.pos().y())
    #         QMouseEvent.accept()
    #     # elif Qt.LeftButton and self._move_drag:
    #     #     self.move(QMouseEvent.globalPos() - self.move_DragPosition)
    #     #     QMouseEvent.accept()

    # def mouseReleaseEvent(self, event):
    #     if self.mPos:
    #         self.mPos = None
    #         self.setCursor(Qt.ArrowCursor)
    #         event.accept()
    #     return

    #     self._move_drag = False
    #     self._corner_drag = False
    #     self._bottom_drag = False
    #     self._right_drag = False
    #     self.setCursor(Qt.ArrowCursor)


class TextEdit(QTextEdit):
    def __init__(self,parent=None):
        super(TextEdit,self).__init__(parent=None)

    def keyPressEvent(self,event):
        if event.key() == Qt.Key_Tab:
            tc = self.textCursor()
            tc.insertText("    ")
            return
        return QTextEdit.keyPressEvent(self,event)

class PlainTextEdit(QPlainTextEdit):
    onSave = lambda : None
    def __init__(self,parent=None):
        super(QPlainTextEdit,self).__init__(parent=None)
        self.keyControlPressed = False

    def keyPressEvent(self,event):
        if event.key() == Qt.Key_Control:
            self.keyControlPressed = True
            return
        elif event.key() == Qt.Key_Tab:
            tc = self.textCursor()
            tc.insertText("    ")
            event.accept()
            return
        elif event.key() == Qt.Key_S:
            self.onSave()
        return QPlainTextEdit.keyPressEvent(self,event)

    def onKeyReleaseEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.keyControlPressed = False

class _ListView(QListView):
    focusout = pyqtSignal()

    def event(self, e):
        if type(e) == QHideEvent:
            # focusOutEvent event not triggered in Linux, maybe a bug?
            # so use hide event instead
            self.focusout.emit()
        return super().event(e)

    # def focusOutEvent(self, event):
    #     self.focusout.emit()

class _Combobox(QComboBox):
    clicked = pyqtSignal()
    listviewFocusout = pyqtSignal()
    def __init__(self):
        super().__init__()
        listView = _ListView()
        listView.executeDelayedItemsLayout()
        listView.focusout.connect(lambda: self.listviewFocusout.emit())
        self.setView(listView)

    def mouseReleaseEvent(self, QMouseEvent):
        self.showItems()

    def showPopup(self):
        # self.popupAboutToBeShown.emit()
        # prevent show popup, manually call it in mouse release event
        pass

    def _showPopup(self):
        max_w = 0
        for i in range(self.count()):
            w = self.view().sizeHintForColumn(i)
            if w > max_w:
                max_w = w
        self.view().setMinimumWidth(max_w + 50)
        super().showPopup()

    def showItems(self):
        self._showPopup()

    def mousePressEvent(self, QMouseEvent):
        self.clicked.emit()


class ButtonCombbox(QWidget):
    activated = pyqtSignal(int)
    currentIndexChanged = pyqtSignal()
    def __init__(self, text="", icon = None, btnClass="smallBtn") -> None:
        super().__init__()
        layout = QHBoxLayout()
        layout.setContentsMargins(2,2,2,2)
        self.setLayout(layout)
        self.button = QPushButton(text)
        self.button.setProperty("class", btnClass)
        if icon:
            utils_ui.setButtonIcon(self.button, icon)
        self.list = _Combobox()
        layout.addWidget(self.button)
        layout.addWidget(self.list)
        self.list.hide()
        self.button.clicked.connect(lambda:self._ctrl("show"))
        def onAvtivated(idx):
            self.activated.emit(idx)
            self._ctrl("hide")
        self.list.activated.connect(lambda idx: onAvtivated(idx))
        self.list.currentIndexChanged.connect(lambda:self.currentIndexChanged.emit())
        self.list.listviewFocusout.connect(self._listFocusout)

    def _ctrl(self, cmd):
        if cmd == "show":
            if self.list.count() == 0:
                return
            self.button.hide()
            self.list.show()
            self.list.showItems()
            self.list.setFocus()
        elif cmd == "hide":
            self.list.hide()
            self.button.show()

    def _listFocusout(self):
        self._ctrl("hide")

    def addItem(self, item):
        self.list.addItem(item)

    def insertItem(self, idx, item):
        self.list.insertItem(idx, item)

    def count(self):
        return self.list.count()

    def findText(self, text):
        return self.list.findText(text)

    def currentIndex(self):
        return self.list.currentIndex()

    def setCurrentIndex(self, idx):
        self.list.setCurrentIndex(idx)

    def currentText(self):
        return self.list.currentText()

class statusBar(QWidget):
    updateUiSignal = pyqtSignal(str, str)
    def __init__(self, rxTxCount = False):
        super().__init__()
        self.rxTxCount = rxTxCount
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.msg = QLabel()
        layout.addWidget(self.msg)
        if rxTxCount:
            self.rx = QLabel('{}({}): {}'.format(_("Received"), _("bytes"), 0))
            self.tx = QLabel('{}({}): {}'.format(_("Sent"), _("bytes"), 0))
            layout.addWidget(self.tx)
            layout.addWidget(self.rx)
        self.updateUiSignal.connect(self._updateUi)
        self.setProperty("class", "statusBar")
        self.rxCount = 0
        self.txCount = 0

    def _updateUi(self, updateType, msg):
        if updateType == "rx":
            self.rx.setText(msg)
        elif updateType == "tx":
            self.tx.setText(msg)
        elif updateType == "msg":
            self.msg.setText(msg)

    def addRx(self, count):
        if not self.rxTxCount:
            return
        self.rxCount += count
        msg = '{}({}): {}'.format(_("Received"), _("bytes"), self.rxCount)
        self.updateUiSignal.emit("rx", msg)

    def addTx(self, count):
        if not self.rxTxCount:
            return
        self.txCount += count
        msg = '{}({}): {}'.format(_("Sent"), _("bytes"), self.txCount)
        self.updateUiSignal.emit("tx", msg)

    def clear(self):
        self.rxCount = 0
        self.txCount = 0
        self.addTx(0)
        self.addRx(0)

    def setMsg(self, level, msg):
        if level == "info":
            color = "#008200"
        elif level == "warning":
            color = "#fb8c00"
        elif level == "error":
            color = "#f44336"
        else:
            color = "#008200"
        msg = '<font color={}>{}</font>'.format(color, msg)
        self.updateUiSignal.emit("msg", msg)

class HelpWidget(QWidget, CustomTitleBarWindowMixin):
    closed = pyqtSignal()
    def __init__(self, pluginsHelp:dict, parent=None, icon=None):
        QWidget.__init__(self, parent)
        CustomTitleBarWindowMixin.__init__(self, init=True, title=_("Help"))
        self.resize(800, 700)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.contentWidget.setLayout(layout)
        # self.setLayout(layout)
        itemList = QListWidget()
        itemList.setProperty("class", "helpList")
        tabs = QStackedWidget()
        layout.addWidget(itemList)
        layout.addWidget(tabs)
        layout.setStretch(0, 1)
        layout.setStretch(1, 4)
        for name, help in pluginsHelp.items():
            tab = QWidget()
            tab.setProperty("class", "helpTab")
            tabLayout = QVBoxLayout()
            tab.setLayout(tabLayout)
            if type(help) == str:
                tabLabel = ScrollLabel(help)
            else:
                tabLabel = help
            tabLayout.addWidget(tabLabel)
            tabs.addWidget(tab)
            itemList.addItem(name)
        itemList.setCurrentRow(0)
        itemList.currentRowChanged.connect(lambda idx: tabs.setCurrentIndex(idx))
        self.show()

    def closeEvent(self, event):
        self.closed.emit()
        event.accept()



class EditRemarDialog(QDialog):
    def __init__(self, remark = "", icon=None, shortcut = [], value=None) -> None:
        super().__init__()
        self.remark = remark
        self.icon = icon
        self.ok = False
        self.settingShortcut = False
        self.shortcut = shortcut
        self.value = value

        layout = QGridLayout()
        self.setLayout(layout)
        layout.addWidget(QLabel(_("Input remark")), 0, 0, 1, 1)
        self.remarkInput = QLineEdit(self.remark)
        if self.value is not None:
            self.valueInput = QLineEdit(self.value)
        self.iconBtn = QPushButton(self.remark)
        if self.icon:
            self.iconBtn.setIcon(qta.icon(self.icon, color="white"))
        if self.shortcut:
            name = "+".join([str(name) for v,name in self.shortcut])
        else:
            name = _("Record")
        self.shortcutBtn = QPushButton(name)
        self.shortcutBtn.setFocusPolicy(Qt.NoFocus)
        layout.addWidget(self.remarkInput, 0, 1, 1, 1)
        if self.value is not None:
            layout.addWidget(QLabel(_("Input value")), 1, 0, 1, 1)
            layout.addWidget(self.valueInput, 1, 1, 1, 1)
        layout.addWidget(QLabel(_("Select icon")), 2, 0, 1, 1)
        layout.addWidget(self.iconBtn, 2, 1, 1, 1)
        layout.addWidget(QLabel(_("Shortcut")), 3, 0, 1, 1)
        layout.addWidget(self.shortcutBtn, 3, 1, 1, 1)
        self.shortcutHint = QLabel(_("Press key to record, or click Cancel"))
        self.shortcutHint.hide()
        layout.addWidget(self.shortcutHint, 4, 0, 1, 2)
        self.okBtn = QPushButton(_("OK"))
        self.cancelBtn = QPushButton(_("Cancel"))
        layout.addWidget(self.okBtn, 5, 0, 1, 1)
        layout.addWidget(self.cancelBtn, 5, 1, 1, 1)

        def ok():
            self.ok = True
            self.close()
        self.okBtn.clicked.connect(lambda : ok())
        self.cancelBtn.clicked.connect(lambda : self.close())
        def updateRemark(text):
            self.remark = text
            self.iconBtn.setText(self.remark)
        self.remarkInput.textChanged.connect(updateRemark)
        self.iconBtn.clicked.connect(lambda: self.selectIcon())
        self.shortcutBtn.clicked.connect(self.setShortcut)

    def selectIcon(self):
        self.icon = selectIcon(parent = self, title = _("Select icon"), btnName = _("OK"), color = utils_ui.getStyleVar("iconSelectorColor"))
        if self.icon:
            self.iconBtn.setIcon(qta.icon(self.icon, color="white"))
        else:
            self.iconBtn.setIcon(QIcon())

    def exec(self):
        super().exec()
        if self.value is not None:
            self.value = self.valueInput.text()
            return self.ok, self.remark, self.value, self.icon, self.shortcut
        return self.ok, self.remark, self.icon, self.shortcut

    def setShortcut(self):
        if not self.settingShortcut:
            self.onRecordShortcut()
        else:
            self.onRecordShortcutEnd()

    def onRecordShortcut(self):
        self.shortcut = []
        self.remarkInput.setEnabled(False)
        self.valueInput.setEnabled(False)
        self.iconBtn.setEnabled(False)
        self.okBtn.setEnabled(False)
        self.cancelBtn.setEnabled(False)
        self.shortcutBtn.setText(_("Cancel"))
        self.shortcutHint.show()
        self.settingShortcut = True
        self.shortcutBtn.setProperty("class", "deleteBtn")
        self.updateStyle(self.shortcutBtn)

    def onRecordShortcutEnd(self, setOk=False):
        self.remarkInput.setEnabled(True)
        self.valueInput.setEnabled(True)
        self.iconBtn.setEnabled(True)
        self.okBtn.setEnabled(True)
        self.cancelBtn.setEnabled(True)
        if not setOk:
            self.shortcutBtn.setText(_("Record"))
            self.shortcut = []
        self.shortcutHint.hide()
        self.settingShortcut = False
        self.shortcutBtn.setProperty("class", "")
        self.updateStyle(self.shortcutBtn)

    def keyPressEvent(self, event):
        if not self.settingShortcut:
            return
        key = event.key()
        name = QKeySequence(key).toString()
        if key == Qt.Key_Control:
            name = "Ctrl"
        elif key == Qt.Key_Shift:
            name = "Shift"
        elif key == Qt.Key_Alt:
            name = "Alt"
        elif key == Qt.Key_Super_L:
            name = "Super_L"
        elif key == Qt.Key_Super_R:
            name = "Super_R"
        self.shortcut.append((key, name))
        keys = "+".join([str(name) for v,name in self.shortcut])
        self.shortcutBtn.setText(keys)

    def keyReleaseEvent(self,event):
        if not self.settingShortcut:
            return
        self.onRecordShortcutEnd(setOk = True)

    def updateStyle(self, widget):
        self.style().unpolish(widget)
        self.style().polish(widget)
        self.update()

class ScrollLabel(QScrollArea):

    # constructor
    def __init__(self, text, *args, **kwargs):
        QScrollArea.__init__(self, *args, **kwargs)

        self.setWidgetResizable(True)
        # making qwidget object
        content = QWidget(self)
        self.setProperty("class", "scrollbar2")
        self.setWidget(content)
        # vertical box layout
        lay = QVBoxLayout(content)
        # creating label
        self.label = QLabel(text, parent = content)
        self.label.setOpenExternalLinks(True)
        self.label.setTextInteractionFlags(Qt.TextSelectableByMouse | Qt.LinksAccessibleByMouse)
        # setting alignment to the text
        self.label.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        # making label multi-line
        self.label.setWordWrap(True)
        # adding label to the layout
        lay.addWidget(self.label)

    def setText(self, text):
        self.label.setText(text)


if __name__ == "__main__":
    import sys
    from PyQt5.QtWidgets import QApplication, QMainWindow, QVBoxLayout

    style = '''
    QWidget {
    }
    .customTilebarWindow {
        background-color: white;
    }
    .TitleBar {
        background-color: #0b1722;
        color: white;
    }
    .TitleBar QPushButton {
        border: none;
    }
    .TitleBar .icon{
        margin-left: 5;
    }
    .TitleBar .title{
        margin-left: 5;
        color: white;
    }
    .TitleBar .top{
        margin-left: 5;
        color: white;
        border-radius: 20;
        background-color: #0b1722;
    }
    .TitleBar .top:hover{
        background-color: #273b4e;
    }
    .TitleBar .topActive{
        margin-left: 5;
        color: white;
        border-radius: 20;
        background-color: #273b4e;
    }
    .TitleBar .min{
        background-color: #53c22a;
        color: white;
    }
    .TitleBar .max{
        background-color: #e5bf28;
        color: white;
    }
    .TitleBar .close{
        background-color: #f45952;
        color: white;
    }
    .TitleBar .min:hover {
        background-color: #2ba13e;
    }
    .TitleBar .max:hover {
        background-color: #cf9001;
    }
    .TitleBar .close:hover {
        background-color: #df2f25;
    }
    QLabel {
        background-color: gray;
    }
    '''

    class MainWindow(QMainWindow, CustomTitleBarWindowMixin):
        _padding = 5
        def __init__(self) -> None:
            super(QMainWindow, self).__init__()
            label = QLabel("hello hhhhhhhhhhh")
            layout = QVBoxLayout()
            layout.addWidget(label)
            self.contentWidget.setLayout(layout)
            self.resize(800, 600)
            self.show()

    app = QApplication(sys.argv)
    app.setStyleSheet(style)
    w = MainWindow()
    # w2 = MainWindow()
    eventFilter = EventFilter()
    eventFilter.listenWindow(w)
    # eventFilter.listenWindow(w2)
    app.installEventFilter(eventFilter)
    app.exec_()
    app.removeEventFilter(eventFilter)




try:
    from Combobox import ComboBox
    from i18n import _
    import utils, parameters
except ImportError:
    from COMTool import utils, parameters
    from COMTool.i18n import _
    from COMTool.Combobox import ComboBox

from PyQt5.QtWidgets import QWidget
from PyQt5.QtGui import QIcon
import qtawesome as qta # https://github.com/spyder-ide/qtawesome

_buttonIcons = {}
_skin = "light"

def setSkin(skin):
    global _skin, _buttonIcons

    if skin == _skin:
        return
    delete = []
    for btn in _buttonIcons:
        if type(btn.parent()) == None:
            delete.append(btn)
            continue
        icon, colorVar = _buttonIcons[btn]
        color = parameters.styleForCode[skin][colorVar]
        btn.setIcon(qta.icon(icon, color=color))
    for btn in delete:
        _buttonIcons.pop(btn)
    _skin = skin    

def setButtonIcon(button, icon : str, colorVar = "iconColor"):
    '''
        @colorVar set in parameters.styleForCode
    '''
    global _skin, _buttonIcons

    iconColor = parameters.styleForCode[_skin][colorVar]
    _buttonIcons[button] = [icon, colorVar]
    button.setIcon(qta.icon(icon, color=iconColor))

def clearButtonIcon(button):
    global _skin, _buttonIcons
    if button in _buttonIcons:
        button.setIcon(QIcon())
        _buttonIcons.pop(button)

def getStyleVar(var):
    global _skin, _buttonIcons

    return parameters.styleForCode[_skin][var]

def updateStyle(parent, widget):
    parent.style().unpolish(widget)
    parent.style().polish(widget)
    parent.update()


import sys
try:
    import parameters
    from i18n import _
    import version
except ImportError:
    from COMTool import parameters
    from COMTool.i18n import _
    from COMTool import version
    
import os
import PyQt5.QtCore
from PyQt5.QtWidgets import QWidget, QLabel, QVBoxLayout
import time


def HelpInfo():
        return '''\
<h1 style='color:#009688';margin=10px;>{}</h1><br>
<b style="color:#009688;margin = 5px;">v{}</b><br><br>
{} + {}<br>
{}: {}<br>
{}: {}<br>
<div>
    <div>{}</div>
    <a style="vertical-align: middle;" href="https://neucrack.com">
        <img src="{}" width=109 height=32/></a>
</div>
{}: <b><a style="color:#009688;" href="https://github.com/Neutree/COMTool#license">LGPL-3.0</a></b><br>
{}<br>
{} <b><a style="color:#009688;" href="https://github.com/neutree/COMTool">Github</a></b>, {} <b><a style="color:#009688;" href="https://github.com/Neutree/COMTool/releases"> releases {}</a></b><br>
{} <b><a style="color:#009688;" href="https://github.com/Neutree/COMTool/issues"> issues</a></b><br>
{}: 566531359 <br><br>
{}: <a style="color:#009688;" href="https://neucrack.com/donate">neucrack.com/donate<br><img src="{}"/> <img src="{}"/></a><br>
'''.format(
    parameters.appName,
    version.__version__,
    '{}{}.{}'.format(sys.implementation.name, sys.implementation.version.major, sys.implementation.version.minor),
    'PyQt{}'.format(PyQt5.QtCore.QT_VERSION_STR),
    _("Config path"),
    parameters.configFilePath,
    _("Old config backup in"),
    os.path.dirname(parameters.configFilePath,),
    _('COMTool is a Open source project create by'),
    '{}/{}'.format(parameters.dataPath, parameters.appLogo2),
    _("License"),
    _('Welcome to improve it together and add plugins'),
    _('See more details on'),
    _("and get latest version at"),
    _("page"),
    _("Have problem? see"),
    _("QQ group for plugin development discussion"),
    _("You can buy me half a cup of coffee if this software helpes you"),
    os.path.join(parameters.assetsDir, "donate_wechat.jpg"),
    os.path.join(parameters.assetsDir, "donate_alipay.jpg")
)




from PyQt5.QtCore import QObject, Qt
from PyQt5.QtWidgets import (QApplication, QWidget,QPushButton,QMessageBox,QDesktopWidget,QMainWindow,
                             QVBoxLayout,QHBoxLayout,QGridLayout,QTextEdit,QLabel,QRadioButton,QCheckBox,
                             QLineEdit,QGroupBox,QSplitter,QFileDialog, QScrollArea)
try:
    from Combobox import ComboBox
    from i18n import _
    import utils, parameters
    from conn.base import ConnectionStatus
    from widgets import statusBar
except ImportError:
    from COMTool import utils, parameters
    from COMTool.i18n import _
    from COMTool.Combobox import ComboBox
    from COMTool.conn.base import  ConnectionStatus
    from COMTool.widgets import statusBar

class Plugin_Base(QObject):
    '''
        call sequence:
            set vars like hintSignal, hintSignal
            onInit
            onWidget
            onUiInitDone
            onActive
                send
                onReceived
            onDel
    '''
    # vars set by caller
    isConnected = lambda o: False
    send = lambda o,x,y:None          # send(data_bytes=None, file_path=None, callback=lambda ok,msg:None), can call in UI thread directly
    ctrlConn = lambda o,k,v:None      # call ctrl func of connection
    hintSignal = None               # hintSignal.emit(type(error, warning, info), title, msg)
    reloadWindowSignal = None       # reloadWindowSignal.emit(title, msg, callback(close or not)), reload window to load new configs
    configGlobal = {}
    # other vars
    connParent = "main"      # parent id
    connChilds = []          # children ids
    id = ""
    name = ""

    enabled = False          # user enabled this plugin
    active  = False          # using this plugin

    help = None                # help info, can be str or QWidget

    def __init__(self):
        super().__init__()
        if not self.id:
            raise ValueError(f"var id of Plugin {self} should be set")

    def onInit(self, config):
        '''
            init params, DO NOT take too long time in this func
            @config dict type, just change this var's content,
                               when program exit, this config will be auto save to config file
        '''
        self.config = config
        default = {
            "version": 1,
        }
        for k in default:
            if not k in self.config:
                self.config[k] = default[k]

    def onDel(self):
        pass

    def onConnChanged(self, status:ConnectionStatus, msg:str):
        '''
            call in UI thread, be carefully!!
        '''
        if status == ConnectionStatus.CONNECTED:
            self.statusBar.setMsg("info", '{} {}'.format(_("Connected"), msg))
        elif status == ConnectionStatus.CLOSED:
            self.statusBar.setMsg("info", '{} {}'.format(_("Closed"), msg))
        elif status == ConnectionStatus.CONNECTING:
            self.statusBar.setMsg("info", '{} {}'.format(_("Connecting"), msg))
        elif status == ConnectionStatus.LOSE:
            self.statusBar.setMsg("warning", '{} {}'.format(_("Connection lose"), msg))
        else:
            self.statusBar.setMsg("warning", msg)

    def onWidgetMain(self, parent):
        '''
            main widget, just return a QWidget object
        '''
        raise NotImplementedError()

    def onWidgetSettings(self, parent):
        '''
            setting widget, just return a QWidget object or None
        '''
        return None

    def onWidgetFunctional(self, parent):
        '''
            functional widget, just return a QWidget object or None
        '''
        return None

    def onWidgetStatusBar(self, parent):
        self.statusBar = statusBar(rxTxCount=False)
        return self.statusBar

    def onReceived(self, data : bytes):
        '''
            call in receive thread, not UI thread
        '''
        for plugin in self.connChilds:
            plugin.onReceived(data)

    def sendData(self, data:bytes):
        '''
            send data, chidren call send will invoke this function
            if you send data in this plugin, you can directly call self.send
        '''
        self.send(data)

    def onKeyPressEvent(self, event):
        pass

    def onKeyReleaseEvent(self, event):
        pass

    def onUiInitDone(self):
        '''
            UI init done, you can update your widget here
            this method runs in UI thread, do not block too long
        '''
        pass

    def onActive(self):
        '''
            plugin active
        '''
        pass

    def bindVar(self, uiObj, varObj, varName: str, vtype=None, vErrorMsg="", checkVar=lambda v:v, invert = False, emptyDefault = None):
        objType = type(uiObj)
        if objType == QCheckBox:
            v = uiObj.isChecked()
            varObj[varName] = v if not invert else not v
            return
        elif objType == QLineEdit:
            v = uiObj.text()
            if v == "" and emptyDefault is not None:
                v = emptyDefault
        elif objType == ComboBox:
            varObj[varName] = uiObj.currentText()
            return
        elif objType == QRadioButton:
            v = uiObj.isChecked()
            varObj[varName] = v if not invert else not v
            return
        else:
            raise Exception("not support this object")
        if vtype:
            try:
                v = vtype(v)
            except Exception:
                uiObj.setText(str(varObj[varName]))
                self.hintSignal.emit("error", _("Error"), vErrorMsg)
                return
        try:
            v = checkVar(v)
        except Exception as e:
            self.hintSignal.emit("error", _("Error"), str(e))
            return
        varObj[varName] = v

    def parseSendData(self, data:str, encoding, usrCRLF=False, isHexStr=False, escape=False):
        if not data:
            return b''
        if usrCRLF:
            data = data.replace("\n", "\r\n")
        if isHexStr:
            if usrCRLF:
                data = data.replace("\r\n", " ")
            else:
                data = data.replace("\n", " ")
            data = utils.hex_str_to_bytes(data)
            if data == -1:
                self.hintSignal.emit("error", _("Error"), _("Format error, should be like 00 01 02 03"))
                return b''
        else:
            if not escape:
                data = data.encode(encoding,"ignore")
            else: # '11234abcd\n123你好\r\n\thello\x00\x01\x02'
                try:
                    data = utils.str_to_bytes(data, escape=True, encoding=encoding)
                except Exception as e:
                    self.hintSignal.emit("error", _("Error"), _("Escape is on, but escape error:") + str(e))
                    return b''
        return data

    def decodeReceivedData(self, data:bytes, encoding, isHexStr = False, escape=False):
        if isHexStr:
            data = utils.bytes_to_hex_str(data)
        elif escape:
            data = str(data)[2:-1] # b'1234\x01' => "b'1234\\x01'" =>"1234\\x01"
        else:
            data = data.decode(encoding=encoding, errors="ignore")
        return data




from os import remove
from PyQt5.QtCore import pyqtSignal,Qt, QRect, QMargins, QObject, pyqtSlot, QRegExp
from PyQt5.QtWidgets import (QWidget,QPushButton,QMessageBox,QDesktopWidget,QMainWindow,
                             QVBoxLayout,QHBoxLayout,QGridLayout,QTextEdit,QLabel,QRadioButton,QCheckBox,
                             QLineEdit,QGroupBox,QSplitter,QFileDialog, QScrollArea)
from PyQt5.QtGui import QIcon,QFont,QTextCursor,QPixmap,QColor, QRegExpValidator
try:
    import parameters,helpAbout,autoUpdate
    from Combobox import ComboBox
    import i18n
    from i18n import _
    import version
    import utils
except ImportError:
    from COMTool import parameters,helpAbout,autoUpdate, utils
    from COMTool.Combobox import ComboBox
    from COMTool import i18n
    from COMTool.i18n import _
    from COMTool import version
try:
    from base import COMM, ConnectionStatus
except Exception:
    from .base import COMM, ConnectionStatus
import socket, threading, time, re




class TCP_UDP(COMM):
    '''
        call sequence:
            onInit
            onWidget
            onUiInitDone
                isConnected
                send
            getConfig
    '''
    id = "tcp_udp"
    name = "TCP UDP"
    showSwitchSignal = pyqtSignal(ConnectionStatus)
    updateTargetSignal = pyqtSignal(str)
    updateClientsSignal = pyqtSignal(bool, tuple)
    def onInit(self, config):
        self.conn = None
        self.config = config
        default = {
            "protocol" : "tcp",
            "mode" : "client",
            "target" : ["127.0.0.1:2345", ["127.0.0.1:2345"]],
            "port": 2345,
            "auto_reconnect": False,
            "auto_reconnect_interval": 1.0
        }
        for k in default:
            if not k in self.config:
                self.config[k] = default[k]
        self.widgetConfMap = {
            "protocol" : None,
            "mode" : None,
            "target" : None,
            "port": None,
            "auto_reconnect": None,
            "auto_reconnect_interval": None
        }
        self.isOpened = False
        self.busy = False
        self.status = ConnectionStatus.CLOSED
        self.widget = None
        self.serverModeClientsConns = {
            # "127.0.0.1:76534": conn
        }
        self.serverModeSelectedClient = None # None means all clients, or ip:port string

    def disconnect(self):
        if self.isConnected():
            self.openCloseSerial()

    def onDel(self):
        if self.isConnected():
            self.openCloseSerial()

    def __del__(self):
        try:
            self.status = ConnectionStatus.CLOSED
            if not self.conn is None:
                self.conn.close()
                self.conn = None
        except Exception:
            pass
        time.sleep(0.05)  # wait for child threads, not wait also ok, cause child threads are daemon

    def getConfig(self):
        '''
            get config, dict type
        '''
        return self.config

    def onUiInitDone(self):
        for key in self.config:
            self.setSerialConfig(key, self.widgetConfMap[key], self.config[key])

    def onWidget(self):
        serialSetting = QWidget()
        self.serialSettingsLayout = QGridLayout()
        protocolLabel = QLabel(_("Protocol"))
        self.modeLabel = QLabel(_("Mode"))
        self.targetLabel = QLabel(_("Target"))
        self.targetCombobox = ComboBox()
        self.targetCombobox.setEditable(True)
        self.portLabel = QLabel(_("Port"))
        self.portLabel.hide()
        self.porttEdit = QLineEdit()
        protocolWidget = QWidget()
        modeWidget = QWidget()
        layoutProtocol = QHBoxLayout()
        layoutMode = QHBoxLayout()
        protocolWidget.setLayout(layoutProtocol)
        modeWidget.setLayout(layoutMode)
        self.protoclTcpRadioBtn = QRadioButton("TCP")
        self.protoclUdpRadioBtn = QRadioButton("UDP")
        self.protoclTcpRadioBtn.setChecked(True)
        layoutProtocol.addWidget(self.protoclTcpRadioBtn)
        layoutProtocol.addWidget(self.protoclUdpRadioBtn)
        self.modeClientRadioBtn = QRadioButton("Client")
        self.modeServerRadioBtn = QRadioButton("Server")
        self.modeClientRadioBtn.setChecked(True)
        layoutMode.addWidget(self.modeClientRadioBtn)
        layoutMode.addWidget(self.modeServerRadioBtn)
        self.clientsCombobox = ComboBox()
        self.clientsCombobox.addItem("0 | " + _("All clients"))
        self.disconnetClientBtn = QPushButton(_("Disconnect"))
        self.autoReconnetLable = QLabel(_("Auto reconnect"))
        self.autoReconnect = QCheckBox()
        self.autoReconnectIntervalEdit = QLineEdit("1.0")
        self.serialOpenCloseButton = QPushButton(_("OPEN"))
        self.serialSettingsLayout.addWidget(protocolLabel,0,0)
        self.serialSettingsLayout.addWidget(protocolWidget, 0, 1, 1, 2)
        self.serialSettingsLayout.addWidget(self.modeLabel, 1, 0)
        self.serialSettingsLayout.addWidget(modeWidget, 1, 1, 1, 2)
        self.serialSettingsLayout.addWidget(self.targetLabel, 2, 0)
        self.serialSettingsLayout.addWidget(self.targetCombobox, 2, 1, 1, 2)
        self.serialSettingsLayout.addWidget(self.portLabel, 3, 0)
        self.serialSettingsLayout.addWidget(self.porttEdit, 3, 1, 1, 2)
        self.serialSettingsLayout.addWidget(self.clientsCombobox, 4, 0, 1, 2)
        self.serialSettingsLayout.addWidget(self.disconnetClientBtn, 4, 2, 1, 1)
        self.serialSettingsLayout.addWidget(self.autoReconnetLable, 5, 0, 1, 1)
        self.serialSettingsLayout.addWidget(self.autoReconnect, 5, 1, 1, 1)
        self.serialSettingsLayout.addWidget(self.autoReconnectIntervalEdit, 5, 2, 1, 1)
        self.serialSettingsLayout.addWidget(self.serialOpenCloseButton, 6, 0, 1, 3)
        serialSetting.setLayout(self.serialSettingsLayout)
        self.widgetConfMap["protocol"]       = [self.protoclTcpRadioBtn, self.protoclUdpRadioBtn]
        self.widgetConfMap["mode"]    = self.modeClientRadioBtn
        self.widgetConfMap["target"]    = self.targetCombobox
        self.widgetConfMap["port"]    = self.porttEdit
        self.widgetConfMap["auto_reconnect"] = self.autoReconnect
        self.widgetConfMap["auto_reconnect_interval"] = self.autoReconnectIntervalEdit
        self.initEvet()
        self.widget = serialSetting
        return serialSetting

    def initEvet(self):
        self.serialOpenCloseButton.clicked.connect(self.openCloseSerial)
        self.porttEdit.textChanged.connect(self.onPortChanged)
        self.showSwitchSignal.connect(self.showSwitch)
        self.updateTargetSignal.connect(self.updateTarget)
        self.updateClientsSignal.connect(self.updateClients)
        self.protoclTcpRadioBtn.clicked.connect(lambda: self.changeProtocol("tcp"))
        self.protoclUdpRadioBtn.clicked.connect(lambda: self.changeProtocol("udp"))
        self.modeServerRadioBtn.clicked.connect(lambda: self.changeMode("server"))
        self.modeClientRadioBtn.clicked.connect(lambda: self.changeMode("client"))
        self.clientsCombobox.currentIndexChanged.connect(self.serverModeClientChanged)
        self.disconnetClientBtn.clicked.connect(self.serverModeDisconnectClient)
        self.autoReconnect.stateChanged.connect(lambda x: self.setVar("auto_reconnect", value = x))
        self.autoReconnectIntervalEdit.textChanged.connect(lambda: self.setVar("auto_reconnect_interval"))
        self.targetCombobox.currentTextChanged.connect(self.onTargetChanged)

    def changeProtocol(self, protocol, init=False):
        if init or protocol != self.config["protocol"]:
            if self.isConnected():
                self.openCloseSerial()
            if protocol == "tcp":
                self.modeClientRadioBtn.show()
                self.modeServerRadioBtn.show()
                self.modeLabel.show()
                self.changeMode(self.config["mode"], init=True)
            else:
                self.targetCombobox.show()
                self.targetLabel.show()
                self.porttEdit.show()
                self.portLabel.show()
                self.clientsCombobox.hide()
                self.disconnetClientBtn.hide()
                self.autoReconnect.hide()
                self.autoReconnectIntervalEdit.hide()
                self.autoReconnetLable.hide()
                self.modeClientRadioBtn.hide()
                self.modeServerRadioBtn.hide()
                self.modeLabel.hide()
                self.widget.adjustSize()
            self.config["protocol"] = protocol

    def changeMode(self, mode, init=False):
        if init or mode != self.config["mode"]:
            if self.isConnected():
                self.openCloseSerial()
            if mode == "server":
                self.targetCombobox.hide()
                self.targetLabel.hide()
                self.porttEdit.show()
                self.portLabel.show()
                self.clientsCombobox.show()
                self.disconnetClientBtn.show()
                self.autoReconnect.hide()
                self.autoReconnectIntervalEdit.hide()
                self.autoReconnetLable.hide()
            else:
                self.targetCombobox.show()
                self.targetLabel.show()
                self.porttEdit.hide()
                self.portLabel.hide()
                self.clientsCombobox.hide()
                self.disconnetClientBtn.hide()
                self.autoReconnect.show()
                self.autoReconnectIntervalEdit.show()
                self.autoReconnetLable.show()
            self.widget.adjustSize()
            self.config["mode"] = mode

    def onTargetChanged(self):
        text = self.targetCombobox.currentText()
        self.config["target"][0] = text

    def updateTarget(self, new):
        idx = self.targetCombobox.findText(new)
        if idx < 0:
            self.targetCombobox.addItem(new)
            self.config["target"][1].append(new)
        self.targetCombobox.setEditText(new)
        self.config["target"][0] = new

    def updateClients(self, add:bool, addr:tuple):
        host, port = addr
        if add:
            self.clientsCombobox.addItem(f'{host}:{port}')
        else:
            idx = self.clientsCombobox.findText(f'{host}:{port}')
            if idx > 0:
                self.clientsCombobox.removeItem(idx)
        self.clientsCombobox.setItemText(0, "{} | ".format(self.clientsCombobox.count() - 1) + _("All clients"))

    def serverModeClientChanged(self):
        if self.clientsCombobox.currentIndex() == 0:
            self.serverModeSelectedClient = None
        else:
            self.serverModeSelectedClient = self.clientsCombobox.currentText()

    def serverModeDisconnectClient(self):
        if not self.serverModeSelectedClient:
            for addr, conn in self.serverModeClientsConns.items():
                try:
                    conn.close()
                except Exception:
                    pass
        else:
            conn = self.serverModeClientsConns[self.serverModeSelectedClient]
            try:
                conn.close()
            except Exception:
                pass

    def onPortChanged(self):
        text = self.porttEdit.text()
        while 1:
            try:
                port = int(text)
                self.config['port'] = port
                break
            except Exception:
                text = text[:-1]
                break
        self.porttEdit.setText(text)

    def onSerialConfigChanged(self, conf_type, obj, value_type, caller=""):
        pass

    def setSerialConfig(self, conf_type, obj, value):
        if conf_type == "protocol":
            if value == "tcp":
                obj[0].setChecked(True)
                obj[1].setChecked(False)
            else:
                obj[0].setChecked(False)
                obj[1].setChecked(True)
        elif conf_type == "mode":
            if value == "client":
                obj.setChecked(True)
                self.changeMode("client", init=True)
            else:
                obj.setChecked(False)
                self.modeServerRadioBtn.setChecked(True)
                self.changeMode("server", init=True)
        elif conf_type == "target":
            for i, target in enumerate(self.config["target"][1]):
                self.targetCombobox.addItem(target)
            self.targetCombobox.setCurrentText(self.config["target"][0])
        elif conf_type == "port":
            obj.setText(str(value))
        elif conf_type == "auto_reconnect":
            obj.setChecked(value)
        elif conf_type == "auto_reconnect_interval":
            obj.setText("%.3f" %(value))

    def setVar(self, key, value = None):
        if key == "auto_reconnect":
            self.config[key] = value
        elif key == "auto_reconnect_interval":
            text = self.autoReconnectIntervalEdit.text()
            try:
                interval = float(text)
                self.config[key] = interval
            except Exception:
                text = "".join(re.findall('[\d\.]*', text))
                self.autoReconnectIntervalEdit.setText(text)

    def openCloseSerial(self):
        if self.busy:
            return
        self.busy = True
        if self.serialOpenCloseButton.text() == _("OPEN"):
            self.isOpened = False
        else:
            self.isOpened = True
        t = threading.Thread(target=self.openCloseSerialProcess)
        t.setDaemon(True)
        t.start()

    def openCloseSerialProcess(self):
        if self.isOpened:
            print("-- disconnect")
            try:
                # set status first to prevent auto reconnect
                self.status = ConnectionStatus.CLOSED
                for addr, conn in self.serverModeClientsConns.items():
                    try:
                        print(f"-- close client {addr} conn")
                        conn.close()
                    except Exception as e:
                        print(f"close client {addr} conn error: ", e)
                        pass
                if not self.conn is None:
                    time.sleep(0.1) # wait receive thread exit
                    try:
                        self.conn.close()
                    except Exception as e:
                        print("close conn error: ", e)
                        pass
                    self.conn = None
            except Exception as e:
                print("openCloseSerialProcess", e)
                pass
            self.onConnectionStatus.emit(self.status, "")
            self.showSwitchSignal.emit(self.status)
        else:
            try:
                if self.config["protocol"] == "tcp":
                    if self.config["mode"] == "client":
                        print("-- connect")
                        target = self.checkTarget(self.config["target"][0])
                        if not target:
                            raise Exception(_("Target error") + ": " + self.config["target"][0])
                        print("-- connect", target)
                        self.onConnectionStatus.emit(ConnectionStatus.CONNECTING, "")
                        self.conn = socket.socket()
                        self.conn.connect(target)
                        self.status = ConnectionStatus.CONNECTED
                        print("-- connect success")
                        self.receiveProcess = threading.Thread(target=self.receiveDataProcess, args=(self.conn, ))
                        self.receiveProcess.setDaemon(True)
                        self.receiveProcess.start()
                    else:
                        print("-- server mode, wait client connect")
                        self.conn = socket.socket()
                        self.conn.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                        self.conn.bind(("0.0.0.0", self.config["port"]))
                        self.conn.listen(100)
                        self.status = ConnectionStatus.CONNECTED
                        self.waitClentsProcess = threading.Thread(target=self.waitClientsProcess)
                        self.waitClentsProcess.setDaemon(True)
                        self.waitClentsProcess.start()
                else:
                    print("-- UPD protocol")
                    self.conn = socket.socket(type=socket.SOCK_DGRAM)
                    self.conn.bind(("0.0.0.0", self.config["port"]))
                    self.status = ConnectionStatus.CONNECTED
                    self.receiveProcess = threading.Thread(target=self.receiveDataProcess, args=(self.conn, ))
                    self.receiveProcess.setDaemon(True)
                    self.receiveProcess.start()
                self.onConnectionStatus.emit(self.status, "")
                self.showSwitchSignal.emit(self.status)
            except Exception as e:
                print("----", e)
                try:
                    self.conn.close()
                    self.conn = None
                except Exception:
                    pass
                msg = _("Connect Failed") +"\n"+ str(e)
                self.hintSignal.emit("error", _("Error"), msg)
                self.status = ConnectionStatus.CLOSED
                self.onConnectionStatus.emit(self.status, msg)
                self.showSwitchSignal.emit(self.status)
        self.busy = False

    def checkTarget(self, target):
        if not target:
            return None
        host = target
        port = 80
        target = target.replace("：", ":")
        if target.endswith(":"):
            target = target[:-1]
        _host = re.match('http(.*)://(.*)', target)
        if _host:
            s, target = _host.groups()
            host = target
        _host = re.match('(.*):(\d*)', target)
        if _host:
            host, port = _host.groups()
            port = int(port)
        if host.endswith("/"):
            host = host[:-1]
        target = (host, port)
        target_str = f'{host}:{port}'
        self.updateTargetSignal.emit(target_str)
        return target

    # @pyqtSlot(ConnectionStatus)
    def showSwitch(self, status):
        if status == ConnectionStatus.CLOSED:
            self.serialOpenCloseButton.setText(_("OPEN"))
            self.serialOpenCloseButton.setProperty("class", "")
        elif status == ConnectionStatus.CONNECTED:
            self.serialOpenCloseButton.setText(_("CLOSE"))
            self.serialOpenCloseButton.setProperty("class", "")
        else:
            self.serialOpenCloseButton.setText(_("CLOSE"))
            self.serialOpenCloseButton.setProperty("class", "warning")
        self.updateStyle(self.serialOpenCloseButton)

    def updateStyle(self, widget):
        self.widget.style().unpolish(widget)
        self.widget.style().polish(widget)
        self.widget.update()

    def waitClientsProcess(self):
        print("-- wait for client connect")
        while self.status != ConnectionStatus.CLOSED:
            # accept connection with non-blocking to check close status
            self.conn.settimeout(0.06)
            try:
                conn, addr = self.conn.accept()
            except socket.timeout:
                continue
            except Exception as e:
                if self.status != ConnectionStatus.CLOSED:
                    print("-- accept connection fail:", str(e))
                continue
            print("-- client connected, ip:", addr)
            addr_str = f'{addr[0]}:{addr[1]}'
            self.updateClientsSignal.emit(True, addr)
            self.onConnectionStatus.emit(ConnectionStatus.CONNECTED, _("Client connected:") + " " + addr_str)
            self.serverModeClientsConns[addr_str] = conn
            t = threading.Thread(target=self.receiveDataProcess, args=(conn, addr))
            t.setDaemon(True)
            t.start()
            print("-- wait for client connect")
        print("-- wait connection thread exit")


    def receiveDataProcess(self, conn, remote_addr:tuple = None):
        waitingReconnect = False
        buffer = b''
        t = 0
        conn.settimeout(0.1)
        protocolIsTcp = self.config["protocol"] == "tcp"
        modeIsServer = self.config["mode"] == "server"
        remoteStr = ""
        if remote_addr:
            remoteStr = f'{remote_addr[0]}:{remote_addr[1]}'
        while self.status != ConnectionStatus.CLOSED:
            if waitingReconnect:
                try:
                    target = self.checkTarget(self.config["target"][0])
                    if not target:
                        raise Exception(_("Target error") + ": " + self.config["target"][0])
                    self.onConnectionStatus.emit(ConnectionStatus.CONNECTING, "")
                    conn = socket.socket()
                    conn.connect(target)
                    conn.settimeout(0.1)
                    self.conn = conn
                    print("-- reconnect")
                    waitingReconnect = False
                    self.onConnectionStatus.emit(ConnectionStatus.CONNECTED, _("Reconnected"))
                    self.showSwitchSignal.emit(ConnectionStatus.CONNECTED)
                    continue
                except Exception as e:
                    pass
                time.sleep(self.config["auto_reconnect_interval"])
                continue
            try:
                # length = max(1, self.conn.in_waiting)
                flush = True
                try:
                    if protocolIsTcp:
                        data = conn.recv(4096)
                        # ignore not selected target's msg
                        if modeIsServer and self.serverModeSelectedClient and (remoteStr != self.serverModeSelectedClient):
                            data = None
                    else:
                        data, target = conn.recvfrom(4096)
                    if data == b'': # closed by peer(peer send FIN, now we can close this connection)
                        if buffer:
                            self.onReceived(buffer)
                        raise Exception(_("Closed by peer"))
                except socket.timeout:
                    data = None
                if data:
                    if len(data) > 4096:
                        flush = False
                    t = time.time()
                    # if length == 1 and not buffer: # just start receive
                    #     buffer += data
                    #     continue
                    buffer += data
                if flush or (buffer and (time.time() - t > 0.001)): # no new data in 0.1ms
                    try:
                        self.onReceived(buffer)
                    except Exception as e:
                        print("-- error in onReceived callback:", e)
                    buffer = b''
            except Exception as e:
                print("-- recv error:", e, type(e))
                if modeIsServer or not self.config["auto_reconnect"]:
                    over = False
                    if protocolIsTcp and modeIsServer:
                        self.onConnectionStatus.emit(ConnectionStatus.CLOSED, _("Connection") + f' {remote_addr[0]}:{remote_addr[1]} ' + _("closed!"))
                        over = True
                    else:
                        self.status = ConnectionStatus.CLOSED
                        self.onConnectionStatus.emit(self.status, _("Connection closed!") + " " + str(e))
                        self.showSwitchSignal.emit(self.status)
                    try:
                        conn.close()
                    except Exception:
                        pass
                    if over:
                        break
                elif (self.status != ConnectionStatus.CLOSED):
                    # close as fast as we can to release port
                    try:
                        conn.close()
                    except Exception:
                        pass
                    waitingReconnect = True
                    self.onConnectionStatus.emit(ConnectionStatus.LOSE, _("Connection lose!"))
                    self.showSwitchSignal.emit(ConnectionStatus.LOSE)
                    time.sleep(self.config["auto_reconnect_interval"])
        # server mode remove client
        remote_str = f'{remote_addr[0]}:{remote_addr[1]}'
        if remote_str in self.serverModeClientsConns:
            print(f"-- client {remote_str} disconnect")
            self.updateClientsSignal.emit(False, remote_addr)
            if remote_str == self.serverModeSelectedClient:
                self.serverModeSelectedClient = None
            self.serverModeClientsConns.pop(remote_str)
        print("-- receiveDataProcess exit")

    def send(self, data : bytes):
        if self.conn:
            if self.config["protocol"] == "tcp":
                if self.config["mode"] == "client":
                    self.conn.sendall(data)
                else:
                    if not self.serverModeSelectedClient:
                        for addr, conn in self.serverModeClientsConns.items():
                            conn.sendall(data)
                    else:
                        self.serverModeClientsConns[self.serverModeSelectedClient].sendall(data)
            else:
                target = self.checkTarget(self.config["target"][0])
                if not target:
                    self.hintSignal.emit("error", _("Target error"), _("Target error") + ": " + self.config["target"])
                self.conn.sendto(data, target)

    def isConnected(self):
        return self.status == ConnectionStatus.CONNECTED




import sys

from qtpy import QtCore, QtGui, QtWidgets

import qtawesome


# TODO: Set icon colour and copy code with color kwarg

VIEW_COLUMNS = 5
AUTO_SEARCH_TIMEOUT = 500
ALL_COLLECTIONS = 'All'


class IconBrowser(QtWidgets.QDialog):
    """
    A small browser window that allows the user to search through all icons from
    the available version of QtAwesome.  You can also copy the name and python
    code for the currently selected icon.
    """

    def __init__(self, parent = None,
                       dialog = False,
                       title = 'QtAwesome Icon Browser',
                       btnOkName = 'OK',
                       btnCopyName = 'Copy Name',
                       color = "gray"
                ):
        super().__init__(parent)

        self.dialog = dialog
        self.selected = ""
        self.color = color

        self.setMinimumSize(400, 600)
        self.setWindowTitle(title)
        if dialog:
            self.setWindowModality(QtCore.Qt.WindowModal)

        qtawesome._instance()
        fontMaps = qtawesome._resource['iconic'].charmap

        iconNames = []
        for fontCollection, fontData in fontMaps.items():
            for iconName in fontData:
                iconNames.append('%s.%s' % (fontCollection, iconName))

        self._filterTimer = QtCore.QTimer(self)
        self._filterTimer.setSingleShot(True)
        self._filterTimer.setInterval(AUTO_SEARCH_TIMEOUT)
        self._filterTimer.timeout.connect(self._updateFilter)

        model = IconModel(self.color)
        model.setStringList(sorted(iconNames))

        self._proxyModel = QtCore.QSortFilterProxyModel()
        self._proxyModel.setSourceModel(model)
        self._proxyModel.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)

        self._listView = IconListView(self)
        self._listView.setUniformItemSizes(True)
        self._listView.setViewMode(QtWidgets.QListView.IconMode)
        self._listView.setModel(self._proxyModel)
        self._listView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        if dialog:
            self._listView.doubleClicked.connect(self._selectIconText)
        else:
            self._listView.doubleClicked.connect(self._copyIconText)
        self._listView.selectionModel().selectionChanged.connect(self._updateNameField)

        self._lineEdit = QtWidgets.QLineEdit(self)
        self._lineEdit.setAlignment(QtCore.Qt.AlignCenter)
        self._lineEdit.textChanged.connect(self._triggerDelayedUpdate)
        self._lineEdit.returnPressed.connect(self._triggerImmediateUpdate)

        self._comboBox = QtWidgets.QComboBox(self)
        self._comboBox.setMinimumWidth(75)
        self._comboBox.currentIndexChanged.connect(self._triggerImmediateUpdate)
        self._comboBox.addItems([ALL_COLLECTIONS] + sorted(fontMaps.keys()))

        lyt = QtWidgets.QHBoxLayout()
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.addWidget(self._comboBox)
        lyt.addWidget(self._lineEdit)
        self._combo_style = QtWidgets.QComboBox(self)
        self._combo_style.addItems([
            qtawesome.styles.DEFAULT_DARK_PALETTE,
            qtawesome.styles.DEFAULT_LIGHT_PALETTE])
        self._combo_style.currentTextChanged.connect(self._updateStyle)
        lyt.addWidget(self._combo_style)

        searchBarFrame = QtWidgets.QFrame(self)
        searchBarFrame.setLayout(lyt)

        self._nameField = QtWidgets.QLineEdit(self)
        self._nameField.setAlignment(QtCore.Qt.AlignCenter)
        self._nameField.setReadOnly(True)


        if self.dialog:
            self._copyButton = QtWidgets.QPushButton(btnOkName, self)
            self._copyButton.clicked.connect(self._selectIconText)
        else:
            self._copyButton = QtWidgets.QPushButton(btnCopyName, self)
            self._copyButton.clicked.connect(self._copyIconText)

        lyt = QtWidgets.QVBoxLayout()
        lyt.addWidget(searchBarFrame)
        lyt.addWidget(self._listView)
        lyt.addWidget(self._nameField)
        lyt.addWidget(self._copyButton)

        self.setLayout(lyt)

        self.setTabOrder(self._comboBox, self._lineEdit)
        self.setTabOrder(self._lineEdit, self._combo_style)
        self.setTabOrder(self._combo_style, self._listView)
        self.setTabOrder(self._listView, self._nameField)
        self.setTabOrder(self._nameField, self._copyButton)
        self.setTabOrder(self._copyButton, self._comboBox)

        QtWidgets.QShortcut(
            QtGui.QKeySequence(QtCore.Qt.Key_Return),
            self,
            self._copyIconText,
        )
        QtWidgets.QShortcut(
            QtGui.QKeySequence("Ctrl+F"),
            self,
            self._lineEdit.setFocus,
        )

        self._lineEdit.setFocus()

        geo = self.geometry()

        # QApplication.desktop() has been removed in Qt 6.
        # Instead, QGuiApplication.screenAt(QPoint) is supported
        # in Qt 5.10 or later.
        try:
            screen = QtGui.QGuiApplication.screenAt(QtGui.QCursor.pos())
            centerPoint = screen.geometry().center()
        except AttributeError:
            desktop = QtWidgets.QApplication.desktop()
            screen = desktop.screenNumber(desktop.cursor().pos())
            centerPoint = desktop.screenGeometry(screen).center()

        geo.moveCenter(centerPoint)
        self.setGeometry(geo)

    def _updateStyle(self, text: str):
        _app = QtWidgets.QApplication.instance()
        if text == qtawesome.styles.DEFAULT_DARK_PALETTE:
            qtawesome.reset_cache()
            qtawesome.dark(_app)
        else:
            qtawesome.reset_cache()
            qtawesome.light(_app)

    def _updateFilter(self):
        """
        Update the string used for filtering in the proxy model with the
        current text from the line edit.
        """
        reString = ""

        group = self._comboBox.currentText()
        if group != ALL_COLLECTIONS:
            reString += r"^%s\." % group

        searchTerm = self._lineEdit.text()
        if searchTerm:
            reString += ".*%s.*$" % searchTerm

        # QSortFilterProxyModel.setFilterRegExp has been removed in Qt 6.
        # Instead, QSortFilterProxyModel.setFilterRegularExpression is
        # supported in Qt 5.12 or later.
        try:
            self._proxyModel.setFilterRegularExpression(reString)
        except AttributeError:
            self._proxyModel.setFilterRegExp(reString)

    def _triggerDelayedUpdate(self):
        """
        Reset the timer used for committing the search term to the proxy model.
        """
        self._filterTimer.stop()
        self._filterTimer.start()

    def _triggerImmediateUpdate(self):
        """
        Stop the timer used for committing the search term and update the
        proxy model immediately.
        """
        self._filterTimer.stop()
        self._updateFilter()

    def _copyIconText(self):
        """
        Copy the name of the currently selected icon to the clipboard.
        """
        indexes = self._listView.selectedIndexes()
        if not indexes:
            return

        clipboard = QtWidgets.QApplication.instance().clipboard()
        clipboard.setText(indexes[0].data())

    def _selectIconText(self):
        """
        Select this button, close widget and return value
        """
        indexes = self._listView.selectedIndexes()
        if not indexes:
            return

        self.selected = indexes[0].data()
        self.close()

    def exec(self):
        super().exec()
        return self.selected

    def _updateNameField(self):
        """
        Update field to the name of the currently selected icon.
        """
        indexes = self._listView.selectedIndexes()
        if not indexes:
            self._nameField.setText("")
        else:
            self._nameField.setText(indexes[0].data())


class IconListView(QtWidgets.QListView):
    """
    A QListView that scales it's grid size to ensure the same number of
    columns are always drawn.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)

    def resizeEvent(self, event):
        """
        Re-implemented to re-calculate the grid size to provide scaling icons
        Parameters
        ----------
        event : QtCore.QEvent
        """
        width = self.viewport().width() - 30
        # The minus 30 above ensures we don't end up with an item width that
        # can't be drawn the expected number of times across the view without
        # being wrapped. Without this, the view can flicker during resize
        tileWidth = width / VIEW_COLUMNS
        iconWidth = int(tileWidth * 0.8)
        # tileWidth needs to be an integer for setGridSize
        tileWidth = int(tileWidth)

        self.setGridSize(QtCore.QSize(tileWidth, tileWidth))
        self.setIconSize(QtCore.QSize(iconWidth, iconWidth))

        return super().resizeEvent(event)


class IconModel(QtCore.QStringListModel):

    def __init__(self, color):
        super().__init__()
        self.color = color

    def flags(self, index):
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

    def data(self, index, role):
        """
        Re-implemented to return the icon for the current index.
        Parameters
        ----------
        index : QtCore.QModelIndex
        role : int
        Returns
        -------
        Any
        """
        if role == QtCore.Qt.DecorationRole:
            iconString = self.data(index, role=QtCore.Qt.DisplayRole)
            return qtawesome.icon(iconString, color = self.color)
        return super().data(index, role)

def selectIcon(parent = None, title = 'QtAwesome Icon Browser', btnName = 'OK', color = "gray"):
    browser = IconBrowser(parent, dialog = True, title=title, btnOkName= btnName, color=color)
    browser.show()
    selectedIcon = browser.exec()
    return selectedIcon

def run(dialog = False):
    """
    Start the IconBrowser and block until the process exits.
    """
    app = QtWidgets.QApplication([])
    qtawesome.dark(app)

    if dialog:
        return selectIcon()

    browser = IconBrowser()
    browser.show()

    sys.exit(app.exec_())


if __name__ == '__main__':
    dialog = False
    if len(sys.argv) > 1:
        dialog = sys.argv[1].lower() in ["true", "1"]
    ret = run(dialog = dialog)
    print("ret: ", ret)





from datetime import datetime
import binascii
import re

def datetime_format_ms(dt):
    res = dt.strftime("%Y-%m-%d %H:%M:%S")
    return '{}.{:03d}'.format(res, int(round(dt.microsecond/1000)))

def hexlify(bs, sed=b' '):
    tmp = b'%02X' + sed.encode()
    return b''.join([tmp % b for b in bs])

def bytes_to_hex_str(strB : bytes) -> str:
    strHex = binascii.b2a_hex(strB).upper()
    return re.sub(r"(?<=\w)(?=(?:\w\w)+$)", " ", strHex.decode())+" "

def hex_str_to_bytes(hexString : str) -> bytes:
    dataList = hexString.split(" ")
    j = 0
    for i in dataList:
        if len(i) > 2:
            return -1
        elif len(i) == 1:
            dataList[j] = "0" + i
        j += 1
    data = "".join(dataList)
    try:
        data = bytes.fromhex(data)
    except Exception:
        return -1
    return data

def str_to_bytes(data:str, escape=False, encoding="utf-8"):
    if not escape:
        return data.encodeing(encoding)
    final = b""
    p = 0
    escapes = {
        "a": (b'\a', 2),
        "b": (b'\b', 2),
        "f": (b'\f', 2),
        "n": (b'\n', 2),
        "r": (b'\r', 2),
        "t": (b'\t', 2),
        "v": (b'\v', 2),
        "\\": (b'\\', 2),
        "\'": (b"'", 2),
        '\"': (b'"', 2),
    }
    octstr = ["0", "1", "2", "3", "4", "5", "6", "7"]
    while 1:
        idx = data[p:].find("\\")
        if idx < 0:
            final += data[p:].encode(encoding, "ignore")
            break
        final += data[p : p + idx].encode(encoding, "ignore")
        p += idx
        e = data[p+1]
        if e in escapes:
            r = escapes[e][0]
            p += escapes[e][1]
        elif e == "x": # \x01
            try:
                r = bytes([int(data[p+2 : p+4], base=16)])
                p += 4
            except Exception:
                msg = "Escape is on, but escape error:" + data[p : p+4]
                raise Exception(msg)
        elif e in octstr and len(data) > (p+2) and data[p+2] in octstr: # \dd or \ddd e.g. \001
            try:
                twoOct = False
                if len(data) > (p+3) and data[p+3] in octstr: # \ddd
                    try:
                        r = bytes([int(data[p+1 : p+4], base=8)])
                        p += 4
                    except Exception:
                        twoOct = True
                else:
                    twoOct = True
                if twoOct:
                    r = bytes([int(data[p+1 : p+3], base=8)])
                    p += 3
            except Exception as e:
                msg = "Escape is on, but escape error:" + data[p : p+4]
                raise Exception(msg)
        else:
            r = data[p: p+2].encode(encoding, "ignore")
            p += 2
        final += r
    return final


def can_draw(ucs4cp):
    return 0x2500 <= ucs4cp and ucs4cp <= 0x259F





try:
    import parameters
except ImportError:
    from COMTool import parameters
import os

protocols_dir = os.path.join(parameters.dataPath, "protocols")

default = '''
def decode(data):
    return data

def encode(data):
    return data
'''

defaultProtocols = {
    "default": default,
}
ignoreList = ["maix-smart"]

for file in os.listdir(protocols_dir):
    name, ext = os.path.splitext(file)
    if name in ignoreList:
        continue
    if ext.endswith(".py"):
        with open(os.path.join(protocols_dir, file)) as f:
            code = f.read()
            defaultProtocols[name] = code





import sys, os
path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "..")
sys.path.insert(0, path)

from conn_tcp_udp import TCP_UDP

if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    from base import ConnectionStatus

    app = QApplication(sys.argv)

    conn = TCP_UDP()
    conn.onInit({})

    class Event():
        def __init__(self, conn) -> None:
            self.conn = conn

        def onReceived(self, data):
            print("-- received:", data)
            # self.conn.send(data)

        def onConnection(self, status, msg):
            print("-- onConnection:", status, msg)
            if status == ConnectionStatus.CONNECTED:
                print("== send data")
                self.conn.send('''GET http://example.com HTTP/1.1
Accept-Language: zh-cn
User-Agent: comtool
Host: example.com:80
Connection: close

'''.encode())

        def onHint(self, level, title, msg):
            print(level, title, msg)

    event = Event(conn)

    conn.onReceived = event.onReceived
    conn.onConnectionStatus.connect(event.onConnection)
    conn.hintSignal.connect(event.onHint)
    window = conn.onWidget()
    conn.onUiInitDone()
    window.show()
    ret = app.exec_()




from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QObject
from enum import Enum


class ConnectionStatus(Enum):
    CLOSED = 0
    CONNECTED = 1
    LOSE = 2 
    CONNECTING = 3


class COMM(QObject):
    '''
        call sequence:
            onInit
            onWidget
            onUiInitDone
                isConnected
                send
            getConfig
            onDel
    '''
    onReceived     = lambda self, data:None   # data: bytes
    onConnectionStatus = pyqtSignal(ConnectionStatus, str)    # connected, msg
    hintSignal = pyqtSignal(str, str, str)                    # hintSignal.emit(type(error, warning, info), title, msg)
    configGlobal = {}
    id = ""
    name = ""

    def __init__(self) -> None:
        super().__init__()
        if (not self.id) or not self.name:
            raise ValueError(f"var id of {self} should be set")

    def onInit(self, config):
        '''
            init params, DO NOT take too long time in this func
        '''
        pass

    def onWidget(self):
        '''
            this method runs in UI thread, do not block too long
        '''
        raise NotImplementedError()

    def onUiInitDone(self):
        '''
            UI init done, you can update your widget here
            this method runs in UI thread, do not block too long
        '''

    def send(self, data : bytes):
        raise NotImplementedError()

    def isConnected(self):
        raise NotImplementedError()

    def disconnect(self):
        raise NotImplementedError()

    def getConfig(self):
        '''
            get config, dict type
            this method runs in UI thread, do not block too long
        '''
        return {}

    def ctrl(self, k, v):
        pass

    def onDel(self):
        '''
            del all things and wait all thread to exit
        '''
        pass